! --------------------------------------------------------------------------------------------------------------------
! This file is part of the package EMar, developed to produce the results in the paper "The marriage market, labor 
!    supply and education choice," by P.A. Chiapppori, M. Costa-Dias and C. Meghir. 
! Copyright (c) 2017 Monica Costa Dias
!
! Routines to solve, simulate and calculate model moments used in estimating the parameters driving 
!     in the education and marital choices as well as the market equilibrium prices
! --------------------------------------------------------------------------------------------------------------------


module solutionSpH

use paths
use precision,    only : wp, neginfty, tol
use constants,    only : ng, nH, ntheta, ns, penalty, mina
use parameters,   only : nParamMax, paramWL, paramS, scaleU, prefC0, prefC1, shareM, pth, pthS, pthC, &
                          costS, mxU, peta, betacoef, paramCopyWL, calcDistrTheta, paramCopyS
use SimStructure, only : simLC, nrepsS, nobsLC
use VstartLife,   only : econgain, upsilon, surplus
use IOroutines,   only : writeSimMomSpH, loadDataMomSpH, readMx, writeResults, writeMx
use functions,    only : getH, recoverS, costEduc
use solutionWL,   only : mtxI, solveWL
use momentsWL,    only : getRatio 
use momData,      only : momSpH_t, Psmsf, Psmsf_sim
use stataMod

type(momSpH_t)     :: momSpH                   ! moment structure
real(wp)           :: sortingmx(ng,0:nH,0:nH)  ! matrix describing sorting patterns
integer, parameter :: totM = ng*nH*(nH+1)      ! moments in sorting matrix
integer, parameter :: bN = 500                 ! replications to estimate variance of marital sorting patterns
real(wp)           :: VprobSpH(totM,totM)      ! variance of marital sorting patterns
real(wp)           :: bPth(bN,ng,ntheta), bPthC(bN,ntheta,ntheta,ns,ns), bPthS(bN,ng,ntheta,ns), &
                      bprobSpH(ng,ntheta,ns,0:nH,bN)

contains



    ! ----------------------------------------------------------------------------------------------------------------
    ! Solution and simulation routines
    !

    !
    ! solve WL & SpH problems, then simulate SpH decisions, calculate moments and print moments to file
    !
    subroutine SolveAndSimSpH(printResults, fileResults, printMom, fileMom, checkMkt, Qform)
        implicit none
        logical, intent(in)      :: printResults, printMom
        integer, intent(in)      :: checkMkt
        character(*), intent(in) :: fileMom, fileResults
        real(wp), intent(out)    :: Qform


        ! initialisations: 
        !     solve WL problem, calculate economic gain from marriage, 
        !     calculate sorting patterns and their SEs, load initial conditions from data,
        !     load data moment and random shocks
        call paramCopyWL(paramWL)             ! copy WL parameters into positions
        call SolveWL                          ! solve WL problem
        call calcGains                        ! calculate gains from marriage g(hm,hf)=exp(betacoef*upsilon)
        call calcDist_probSpH                 ! calculate distribution of marital patterns: p(S,pH | g,th)
        call loadDataMomSpH(momSpH)           ! load data moments for SpH
        call InitCondSpH                      ! get initial conditions & shocks (normalised)

        ! solve education choice and marital sorting problems, simulate choices and moments, calculate Qform
        call calcQformSpH(1, 0, checkMkt, 1, 1, Qform)

        ! export data moments & results
        if (printMom==.true.)     call writeSimMomSpH(momSpH, fileMom)
        if (printResults==.true.) call writeResults(765,fileResults,sortingmx)

    end subroutine SolveAndSimSpH


    !
    ! calculate the objective function, QformSpH
    !
    subroutine calcQformSpH(all, sharing, checkMkt, dS, dpH, Qform)
        implicit none
        integer, intent(in)   :: all, sharing, checkMkt, dS, dpH
        real(wp), intent(out) :: Qform


        ! load parameters and simulate marital and education choices
        call paramCopyS(paramS%val(1:paramS%n), paramS%n, econgain)
        call simulateSpH(checkMkt, dS, dpH)
        
        ! ojective function when estimating all parameters
        if (all==1) then
            call SimMomentsSpH(Qform)

        ! ojective function when calculating sharing rule that clears market
        else if (sharing==1) then
            call calcSorting
            Qform = sum( (sortingmx(1,1:nH,1:nH)-sortingmx(2,1:nH,1:nH)) * (sortingmx(1,1:nH,1:nH)-sortingmx(2,1:nH,1:nH)) )
        end if

    end subroutine calcQformSpH



    !
    ! Simulation routine: 
    !   simulates education and marital choices and their distribution given future 
    !   values and present costs/preferences
    ! Allows to:
    !   1. conditional on education, simulate only marital choice (dS=0)
    !   2. conditional on marital choice, simulate only education (dpH=0)
    ! In these cases, the conditioning variable must be set before calling this routine
    ! Also allows to shutdown (or not) infeasible markets - these are markets with 
    !   (unrestricted) sharing rule outside (0,1) interval (checkMkt=1/0, respectively)
    !
    subroutine simulateSpH(checkMkt, dS, dpH)
        implicit none
        integer, intent(in) :: checkMkt, dS, dpH
        real(wp)            :: cS(ns), indU(ns,0:nH), V1(nrepsS,ns), V0(nrepsS,ns), U(ng,nH,0:nH)
        integer             :: i, g, pg, s, optS(2), optSpH(2), optpH(1), opt1d(1), h, ph, th, rep, ps, d, j
        

        ! organise utility parameters for estimation: exp(utility (H,pH) x betacoef)
        U = 0.0_wp
        forall (h=1:nH, ph=0:nH) 
            U(1,h,ph) = mxU(1,h,ph)
            U(2,h,ph) = mxU(2,ph,h)
        end forall


        ! simulate education and marital choices for all individuals
        do i=1,nobsLC
        
            g  = simLC(i)%female+1
            pg = 2-simLC(i)%female
            th = simLC(i)%theta
            forall (s=1:3)        cS(s) = costEduc (g, th, s, simLC(i)%pYres, 0.0_wp)     
            forall (rep=1:nrepsS) simLC(i)%SpH(rep)%probSpH_byK = 0.0_wp

            ! simulate prob (S,pH) given individual characteristics (g,theta,X,kappa)
            do rep=1,nrepsS

                indU = 0.0_wp

                ! value of each option of own education and partner's HC (s,pH) stored in indU
                h = getH(th,1);  indU(1,:) = scaleU(g)*U(g,h,:) + simLC(i)%SpH(rep)%eps(:) - simLC(i)%SpH(rep)%eps(0)
                h = getH(th,2);  indU(2,:) = scaleU(g)*U(g,h,:) - cS(2) - simLC(i)%SpH(rep)%kappa(2) + simLC(i)%SpH(rep)%eps(:) - simLC(i)%SpH(rep)%eps(0)
                h = getH(th,3);  indU(3,:) = scaleU(g)*U(g,h,:) - cS(3) - simLC(i)%SpH(rep)%kappa(3) + simLC(i)%SpH(rep)%eps(:) - simLC(i)%SpH(rep)%eps(0)
                indU(:,0) = indU(:,0) + prefC0(g,th,:)
                do s=1,ns 
                do ps=1,ns 
                    d = abs(ps-s)
                    if (d>0) then
                        indU(s,ps)    = indU(s,ps)+prefC1(g,th,d)
                        indU(s,ps+ns) = indU(s,ps+ns)+prefC1(g,th,d)
                    end if
                end do
                end do

                ! elliminate impossible options (sharing rule outside unit interval) if checkMk=1
                if (checkMkt==1) then
                    do s=1,ns
                    do ph=1,nH
                        h = getH ( th , s )
                        if ( g==1 .and. ( shareM(h,ph)<=0.0101_wp .or. shareM(h,ph)>=0.9899_wp ) ) indU(s,ph) = neginfty 
                        if ( g==2 .and. ( shareM(ph,h)<=0.0101_wp .or. shareM(ph,h)>=0.9899_wp ) ) indU(s,ph) = neginfty 
                    end do
                    end do
                end if

                ! identify optimal choice: either SpH if unconditional, 
                !    or only S or pH if conditional on the other 
                !    (in which case the conditioning var needs to be set in advance)
                if (dS==1 .and. dpH==1) then
                    optSpH = maxloc(indU)
                    simLC(i)%SpH(rep)%s  = optSpH(1)
                    simLC(i)%SpH(rep)%ph = optSpH(2)-1
                else if (dS==0) then
                    opt1d = maxloc( indU(simLC(i)%SpH(rep)%s,:) )
                    simLC(i)%SpH(rep)%ph = opt1d(1)-1
                else if (dpH==0) then
                    opt1d = maxloc( indU(:,simLC(i)%SpH(rep)%ph) )
                    simLC(i)%SpH(rep)%s = opt1d(1)
                end if
                simLC(i)%SpH(rep)%h  = getH ( th , simLC(i)%SpH(rep)%s )

                ! store cost of education
                simLC(i)%SpH(rep)%cS = cS+simLC(i)%SpH(rep)%kappa
                
                ! store values of all options
                simLC(i)%SpH(rep)%VSpH = indU
                
                ! store value of alternative education choices at unconditional optimal marital choice
                optSpH = maxloc(indU)
                do s=1,ns
                    V1(rep,s) = indU(s,optSpH(2)-1)
                end do

                ! store value of alternative education choices at optimal marital choice conditional on education
                do s=1,ns
                    optpH = maxloc(indU(s,:))
                    V0(rep,s) = indU(s,optpH(1)-1)
                    simLC(i)%SpH(rep)%pHbyS(s) = optpH(1)-1
                end do
                    
            end do

            ! calculate probability of education and marital choices
            simLC(i)%probSpH = 0.0_wp
            simLC(i)%probS   = 0.0_wp
            simLC(i)%probSpS = 0.0_wp
            do s=1,ns
            do ph=0,nH
                simLC(i)%probSpH(s,ph) = sum( merge( 1.0_wp, 0.0_wp, simLC(i)%SpH(:)%s==s .and. simLC(i)%SpH(:)%ph==ph ) ) / nrepsS
            end do
            end do
            do s=1,ns
                simLC(i)%probS(s)     = sum( simLC(i)%probSpH(s,:) )
                simLC(i)%probSpS(s,0) = simLC(i)%probSpH(s,0)
                do ph = 1,nH
                    ps = recoverS(ph)
                    simLC(i)%probSpS(s,ps) = simLC(i)%probSpS(s,ps)+simLC(i)%probSpH(s,ph)
                end do
            end do

            ! store expected cost of education
            forall (s=1:ns) simLC(i)%EcS(s) = sum(simLC(i)%SpH(:)%cS(s))/real(nrepsS,wp)
            
            ! store value of education
            forall (s=1:ns) simLC(i)%V0(s) = sum(V0(:,s))/real(nrepsS,wp)
            forall (s=1:ns) simLC(i)%V1(s) = sum(V1(:,s))/real(nrepsS,wp)

        end do
    
    end subroutine simulateSpH




    ! ----------------------------------------------------------------------------------------------------------------
    ! Moments routines
    !

    !
    ! Calculate all moments used to estimate parameters in SpH problem, on simulated data 
    ! Then calculate value of minimising criterion
    !

    subroutine SimMomentsSpH(Qform)
        use nag_g_ib, only : g02daf
        implicit none
        real(wp), intent(out) :: Qform
        integer, parameter    :: nb=3
        integer               :: th, g, s, ps, imom, ph, df, rank, ifail, n, i, rep, inX(nb-1), unit, &
                                 female(nobsLC*nrepsS), inregr(nobsLC*nrepsS)
        real(wp)              :: x1, x2, wgt(nobsLC*nrepsS), rss, se(nb), cov(nb*(nb+1)/2), H(nobsLC*nrepsS), &
                                 Q(nobsLC*nrepsS,nb+1), P(2*nb+nb*nb), wk(2*(nb-1)+nb*nb), res(nobsLC*nrepsS), &
                                 X(nobsLC*nrepsS,nb-1), mxS(nobsLC*nrepsS,ns), eqpenalty, b(nb)
        logical               :: svd


        ! probability(educ S, marry pH | theta, g)
        do g  = 1,ng
        do th = 1,ntheta
        do s  = 1,ns
            imom = 0
            do ph = 0,nH
                imom = imom+1
                x1 = sum( merge(1.0_wp, 0.0_wp, simLC(:)%female==g-1 .and. simLC(:)%theta==th .and. simLC(:)%insampleWL==1 ) * simLC(:)%probSpH(s,ph) )
                x2 = sum( merge(1.0_wp, 0.0_wp, simLC(:)%female==g-1 .and. simLC(:)%theta==th .and. simLC(:)%insampleWL==1 ) )
                call getRatio( x1, x2, momSpH%probSpH(g,th)%dta(imom,s), momSpH%probSpH(g,th)%se(imom,s), momSpH%probSpH(g,th)%sim(imom,s), momSpH%probSpH(g,th)%sediff(imom,s) )
            end do
        end do
        end do
        end do

        ! marital choices by education
        do g  = 1,ng
        do s  = 1,ns
            imom = 0
            do ps = 1,ns
                imom = imom+1
                x1 = sum( merge(1.0_wp, 0.0_wp, simLC(:)%female==g-1 .and. simLC(:)%insampleWL==1 ) * simLC(:)%probSpS(s,ps) )
                x2 = sum( merge(1.0_wp, 0.0_wp, simLC(:)%female==g-1 .and. simLC(:)%insampleWL==1 ) )
                if (g==1) Psmsf_sim(g,s,ps) = x1/x2
                if (g==2) Psmsf_sim(g,ps,s) = x1/x2
            end do
        end do
        end do

        ! probability(educ S| g) 
        do g = 1,ng
        do s = 1,ns
            imom = 1
            x1 = sum( merge(1.0_wp, 0.0_wp, simLC(:)%female==g-1 .and. simLC(:)%insampleWL==1 ) * simLC(:)%probS(s) )
            x2 = sum( merge(1.0_wp, 0.0_wp, simLC(:)%female==g-1 .and. simLC(:)%insampleWL==1 ) )
            call getRatio( x1, x2, momSpH%probS(g)%dta(imom,s), momSpH%probS(g)%se(imom,s), momSpH%probS(g)%sim(imom,s), momSpH%probS(g)%sediff(imom,s) )
        end do
        end do
        

        ! regression: educ S on residual parental income and constant by gender

        ! prepare regressors and dependent variables
        n = 0
        do i = 1,nobsLC
            X(n+1:n+nrepsS,1)    = simLC(i)%pYres
            X(n+1:n+nrepsS,2)    = simLC(i)%pYres*simLC(i)%pYres
            female(n+1:n+nrepsS) = simLC(i)%female
            inregr(n+1:n+nrepsS) = simLC(i)%insampleS
            do rep = 1,nrepsS
                n = n+1
                mxS(n,1) = merge( 1.0_wp, 0.0_wp, simLC(i)%SpH(rep)%s==1 )
                mxS(n,2) = merge( 1.0_wp, 0.0_wp, simLC(i)%SpH(rep)%s==2 )
                mxS(n,3) = merge( 1.0_wp, 0.0_wp, simLC(i)%SpH(rep)%s==3 )
            end do
        end do

        ! run regressions
        n   = nobsLC*nrepsS
        inX = 1
        do g = 1,ng
            do s = 1,ns
                wgt = merge(1.0_wp, 0.0_wp, female==g-1 .and. inregr==1)
                ifail = -1
                call G02DAF( mean='M',   weight='W', n=n,        X=X,     ldX=n,   m=nb-1,                                &
                             isX=inX,    ip=nb,      y=mxS(:,s), wt=wgt,  rss=rss, idf=df, b=momSpH%regrS(g)%sim(1:nb,s), &
                             se=se,      cov=cov,    res=res,    H=H,     Q=Q,     ldQ=n,  svd=svd,                       &
                             irank=rank, P=P,        tol=tol,    wk=wk,   ifail=ifail )
                if (ifail==0) then
                    momSpH%regrS(g)%sediff(1:nb,s) = abs(momSpH%regrS(g)%sim(1:nb,s)-momSpH%regrS(g)%dta(1:nb,s)) / momSpH%regrS(g)%se(1:nb,s)
                else
                    momSpH%regrS(g)%sim(1:nb,s)    = penalty
                    momSpH%regrS(g)%sediff(1:nb,s) = penalty
                end if
            end do
        end do


        ! penalty for distance from market clearing prices
        call calcSorting
        eqpenalty = 10.0_wp * sum( (sortingmx(1,1:nH,1:nH)-sortingmx(2,1:nH,1:nH)) * (sortingmx(1,1:nH,1:nH)-sortingmx(2,1:nH,1:nH)) )


        ! calculate sum of squared differences and objective function
        do g=1,ng
            forall (th=1:ntheta) &
            momSpH%probSpH(g,th)%ssdiff = sum ( ( momSpH%probSpH(g,th)%sediff(1:momSpH%probSpH(g,th)%n,:) )**2 )
            momSpH%probS(g)%ssdiff      = sum ( ( momSpH%probS(g)%sediff(1,:) )**2 )
            momSpH%regrS(g)%ssdiff      = sum ( ( momSpH%regrS(g)%sediff(2:momSpH%regrS(g)%n,:) )**2 )
        end do
        Qform = sum( momSpH%probSpH(:,:)%ssdiff ) + sum( momSpH%probS(:)%ssdiff ) + sum( momSpH%regrS(:)%ssdiff ) + eqpenalty

    end subroutine SimMomentsSpH


    !
    ! calculate sorting patterns from model simulations
    !
    subroutine calcSorting
        implicit none
        integer             :: g, th, s, h, ph, mh, wh
        real(wp)            :: x1, x2
        
        do g  = 1,ng
        do th = 1,ntheta
        do s  = 1,ns
            h  = (th-1)*ns+s
            do ph = 0,nH
                x1 = sum( merge(1.0_wp, 0.0_wp, simLC(:)%female==g-1 .and. simLC(:)%theta==th .and. simLC(:)%insampleWL==1) * simLC(:)%probSpH(s,ph) )
                x2 = sum( merge(1.0_wp, 0.0_wp, simLC(:)%female==g-1 .and. simLC(:)%theta==th .and. simLC(:)%insampleWL==1) )
                mh = merge(h,ph,g==1)
                wh = merge(h,ph,g==2)
                if (x2>tol) then
                    sortingmx(g,mh,wh) = 100.0_wp*pth(g,th)*x1/x2
                else
                    sortingmx(g,mh,wh) = 0.0_wp
                end if
            end do
        end do
        end do
        end do
        
    end subroutine calcSorting




    ! ----------------------------------------------------------------------------------------------------------------
    ! Calculate economic value of marriage and singlehood, conditional on HC
    !

    subroutine calcGains
        implicit none
        integer  :: ithm, ithf, ith, sm, sf, s, ihm, ihf, ih


        ! calculate LT value of marriage (Hm,Hf) exc private gains (driven by PW - mu)
        upsilon = 0.0_wp

        ! expected value of remaining single
        do ith = 1,ntheta
        do s   = 1,ns
            ih = getH(ith,s)
            upsilon(ih,0) = mtxI(mina)%S(1,1,ih,1,1)*peta(1,s) + mtxI(mina)%S(1,2,ih,1,1)*(1.0_wp-peta(1,s))
            upsilon(0,ih) = mtxI(mina)%S(2,1,ih,1,1)*peta(2,s) + mtxI(mina)%S(2,2,ih,1,1)*(1.0_wp-peta(2,s))
        end do
        end do

        ! expected value of marriage (excl marital preferences) and marital surplus
        do ithm = 1,ntheta
        do ithf = 1,ntheta
        do sm   = 1,ns
        do sf   = 1,ns
            ihm     = getH(ithm,sm)
            ihf     = getH(ithf,sf)
            upsilon(ihm,ihf) = mtxI(mina)%C(1,1,ihm,ihf,1,1,1) * peta(1,sm) * peta(2,sf)          + &
                               mtxI(mina)%C(1,2,ihm,ihf,1,1,1) * peta(1,sm) * (1.0_wp-peta(2,sf)) + &
                               mtxI(mina)%C(2,1,ihm,ihf,1,1,1) * (1.0_wp-peta(1,sm)) * peta(2,sf) + &
                               mtxI(mina)%C(2,2,ihm,ihf,1,1,1) * (1.0_wp-peta(1,sm)) * (1.0_wp-peta(2,sf)) 
        end do
        end do
        end do
        end do

        ! economic gain from marriage
        econgain      = exp( upsilon * betacoef ) 
        econgain(0,0) = 0.0_wp
        forall(ihm=1:nH, ihf=1:nH) surplus(ihm,ihf) = econgain(ihm,ihf)-econgain(0,ihf)-econgain(ihm,0)

    end subroutine calcGains




    ! ----------------------------------------------------------------------------------------------------------------
    ! Construct P(S, pH | th, g) from model estimates; then bootstrap its SE and store in
    !    estimates in moment structure
    ! P(S, pH | th, g) needed to estimate of parameters driving choices of education & sorting 
    !
    
    !
    ! calculate distribution of P(S, pH | th, g): wrapper
    !
    subroutine calcDist_probSpH
        implicit none
        real(wp) :: pSpH(ng,ntheta,ns,0:nH)


        ! calculate distribution of education and marital choices and BS st errors
        call calc_ProbSpH(Psmsf, pSpH)
        call bSE_probSpH(paramWL%nToEstimate, VprobSpH, totM)

        ! fill moment structure
        call store_probSpH(pSpH, VprobSpH, totM)
                
    end subroutine calcDist_probSpH


    !
    ! Calculate P(S,pH | g, theta) given set of parameters
    !
    subroutine calc_ProbSpH(Msmsf,pSpH)
        implicit none
        real(wp), intent(in)  :: Msmsf(0:ns,0:ns)
        real(wp), intent(out) :: pSpH(ng,ntheta,ns,0:nH)
        integer               :: s, ss, ith, iths, ihs
        real(wp)              :: massM, massF


        ! Probability of choice (S,pH) for individual (g,th): probSpH(g,th,S,pH)=p(S,pH | g,th)
        pSpH  = 0.0_wp
        massM = sum(Msmsf(1:ns,:))
        massF = sum(Msmsf(:,1:ns))
        
        do ith=1,ntheta
        do s = 1,ns
        
            ! prob choosing educ s and remaining single
            pSpH(1,ith,s,0) = pthS(1,ith,s)*Msmsf(s,0)/(massM*Pth(1,ith))                 ! 
            pSpH(2,ith,s,0) = pthS(2,ith,s)*Msmsf(0,s)/(massF*Pth(2,ith))

            ! prob choosing educ s and marry spouse (ss,iths)
            do ss = 1,ns
            do iths = 1,ntheta
                ihs = getH(iths,ss)
                pSpH(1,ith,s,ihs) = pthC(ith,iths,s,ss)*Msmsf(s,ss)/(massM*Pth(1,ith))
                pSpH(2,ith,s,ihs) = pthC(iths,ith,ss,s)*Msmsf(ss,s)/(massF*Pth(2,ith))
            end do
            end do

        end do
        end do
       
    end subroutine calc_ProbSpH


    !
    ! Bootstrap SE of P(S, pH | th, g) 
    !
    subroutine bSE_probSpH(npar, VpSpH, pN)
        use nag_f_ib, only : f07fdf
        use nag_g_ib, only : g02bxf
        implicit none
        integer, intent(in)   :: npar, pN
        real(wp), intent(out) :: VpSpH(pN,pN)
        real(wp)              :: bParam(npar,bN), bP((ns+1)*(ns+1),bN), par(nParamMax), X(bN, pN), w(1),  &
                                 R(pN,pN), meanX(pN), seX(pN), cdf(bN), bPsmsf(0:ns,0:ns,bN), Psmsf0(0:ns,0:ns)
        integer               :: i, j, n, ifail, g, ith, s, ihs, sm, sf

        
        ! draw BS sample
        call readMx(npar, bN, bParam, mompath//'rd-paramWL.dta', 'r', 1)
        call readMx((ns+1)*(ns+1), bN, bP, mompath//'rd-Psmsf.dta', 'P', 1)
        n=0
        do sf=0,ns
        do sm=0,ns
            n=n+1;  bPsmsf(sm,sf,:) = bP(n,:)
        end do
        end do

        ! store original parameters
        par = paramWL%val           
        Psmsf0 = Psmsf

        ! calculate probSpH for each BS draw
        do i=1,bN
        
            ! read BS draw of parameters
            n = 0
            do j=1,paramWL%n
            if (paramWL%toestimate(j)==.true.) then
                n=n+1;  paramWL%val(j)=bParam(n,i)
            end if
            end do
            Psmsf = bPsmsf(:,:,i)
            call paramCopyWL(paramWL)
            
            ! calculate distribution of education and marital choice
            bPth(i,:,:)      = Pth
            bPthC(i,:,:,:,:) = PthC
            bPthS(i,:,:,:)   = PthS
            call calc_probSpH(Psmsf, bprobSpH(:,:,:,:,i))
            
        end do

        ! recover original set of parameters
        paramWL%val = par
        Psmsf = Psmsf0
        call paramCopyWL(paramWL)
        
        ! calculate covariance matrix of probSpH 
        n = 0
        do g=1,ng
        do ith=1,ntheta
        do s=1,ns
        do ihs=0,nH
            n=n+1;  X(:,n)=bprobSpH(g,ith,s,ihs,:)
        end do
        end do
        end do
        end do
        w = 0
        ifail=0;  call g02bxf( 'U', bN, pN, X, bN, w, meanX, seX, VpSpH, pN, R, ifail)

    end subroutine bSE_probSpH


    !
    ! store distribution of education and marital choices 
    !
    subroutine store_probSpH(pSpH, VpSpH, pN)
        implicit none
        integer, intent(in)  :: pN
        real(wp), intent(in) :: pSpH(ng,ntheta,ns,0:nH), VpSpH(pN,pN)
        integer g, th, ihs, s, imom

        ! store probSpH and its SE
        forall (g=1:ng, th=1:ntheta)
            momSpH%probSpH(g,th)%name     = ''
            momSpH%probSpH(g,th)%dta      = 0.0_wp
            momSpH%probSpH(g,th)%dtaQ     = 0.0_wp
            momSpH%probSpH(g,th)%se       = 0.0_wp
            momSpH%probSpH(g,th)%sim      = 0.0_wp
            momSpH%probSpH(g,th)%sediff   = 0.0_wp
        end forall

        imom = 0
        do g=1,ng
        do th=1,ntheta
        do s=1,ns
        do ihs=0,nH
            imom = imom+1
            if(s==1) write(momSpH%probSpH(g,th)%name(ihs+1), '("(g,th,ph)=(", I1, ",", I1, ",", I1)') g, th, ihs
            momSpH%probSpH(g,th)%dta(ihs+1,s) = pSpH(g,th,s,ihs)
            momSpH%probSpH(g,th)%se(ihs+1,s)  = sqrt(VpSpH(imom,imom))
        end do
        end do
        end do
        end do
        momSpH%probSpH(:,:)%n = nH+1
        
    end subroutine store_probSpH
    


    ! ----------------------------------------------------------------------------------------------------------------
    ! Read data: initial conditions for sorting and education problem
    ! Draw shocks
    !

    subroutine InitCondSpH
        use nag_g_ib,  only : g05kbf, g05lgf
        implicit none
        integer               :: unit, i, g, ifail, igen, iseed(4), rep, s, th, ph
        real(wp),allocatable  :: rvtheta(:,:), x


        ! Read initial conditions and sample from STATA
        unit = 5454
        call openStata(dtaPath//'BHPSsampleSpH.dta', unit)
        nobsLC = nobsStata()
        if (allocated(simLC)) deallocate(simLC);   allocate (simLC(nobsLC))
        call readStata(simLC(:)%female,     'w'     )
        call readStata(simLC(:)%pYres,      'pYres' )
        call readStata(simLC(:)%insampleWL, 'insampleWL' )
        call readStata(simLC(:)%insampleS,  'insampleEduc' )
        call closeOpenStata()         

        ! Draw shocks known at the start of life
        igen     = 0
        iseed    = 0
        iseed(1) = 2576654
        call g05kbf(igen,iseed)

        ! uniform rv for theta (ability)
        igen  = 0
        ifail = 0
        call g05lgf(0.0_wp, 1.0_wp, nobsLC, simLC(:)%rvtheta, igen, iseed, ifail)
        do i=1,nobsLC
            simLC(i)%theta = merge( 1, 2 , simLC(i)%rvtheta <= Pth(simLC(i)%female+1,1) )
        end do

        ! normalised (SE=1) normal shocks in preferences for education (kappa)
        do rep = 1,nrepsS
            simLC(:)%SpH(rep)%kappa(1)=0.0_wp
            do s = 2,ns
                igen  = 0
                ifail = 0
                call g05laf (0.0_wp, 1.0_wp, nobsLC, simLC(:)%SpH(rep)%kappa(s), igen, iseed, ifail)
            end do
        end do

        ! extreme value shocks on preferences for different spouses or remaining single (epsilon)
        do rep = 1,nrepsS
        do ph  = 0,nH
            igen  = 0
            ifail = 0
            call g05lgf(0.0_wp, 1.0_wp, nobsLC, simLC(:)%SpH(rep)%eps(ph), igen, iseed, ifail)
            simLC(:)%SpH(rep)%eps(ph) = -log(-log(simLC(:)%SpH(rep)%eps(ph)))
        end do
        end do

    end subroutine InitCondSpH

end module solutionSpH

