! --------------------------------------------------------------------------------------------------------------------
! This file is part of the package EMar, developed to produce the results in the paper "The marriage market, labor 
!    supply and education choice," by P.A. Chiapppori, M. Costa-Dias and C. Meghir. 
! Copyright (c) 2017 Monica Costa Dias
!
! Routines to estimate the parameters in the post-marriage/education problem, including parameters in the earnings 
!    equation and the preference parameters, as well as the distribution of ability conditional on marital status and
!    the education of both spouses
! --------------------------------------------------------------------------------------------------------------------

module estimationWL

use paths
use precision,     only : wp, infty, neginfty, tol
use constants,     only : ng, ns
use parameters,    only : param_t, paramWL
use momData,       only : momWL_t
use simStructure,  only : nrepsWL
use functions,     only : scaleToRange, realToStr, invMx
use IOroutines,    only : paramSaveWL, PrintEstimates, writeMx, readMx
use simulationWL,  only : SolveAndSimWL
use momentsWL,     only : SimMomentsWL
use statamod

real(wp)         :: QformScale
integer, private :: ncallQA


contains
    

    !-----------------------------------------------------------------------------------------------------------------
    ! Estimation procedures
    !
    
    !
    ! Main routine: Quadratic approximation
    !
    subroutine estimateWL
        use nag_e_ib,  only : e04jcf
        implicit none
        integer                                 :: iuser(1), ifail, ncall, n, ntoestimate, maxIter
        real(wp),dimension(paramWL%nToEstimate) :: b, bl, bu
        real(wp)                                :: Qform, scaledQform, ruser(1)
        type(momWL_t)                           :: mom

        ! Initialisation: reads parameters and computes moments, Qform and scale of Qform, exports info to screen
        QformScale = 0.0_wp
        call getQ_WL(Qform)
        call scaleToRange(Qform, QformScale, scaledQform)         
        write(*,'(/,"    Initial Sum of Squared Differences in Moments: ", F15.6, /)') Qform

        ! estimate objective function
        call QAparameters(paramWL)
        ntoestimate = count(paramWL%toestimate)
        n           = paramWL%n
        maxIter     = 10000
        ncallQA     = 1
        ifail       = -1
        call e04jcf(objectiveCalcQA, n, 2*ntoestimate+1, paramWL%b(1:n), paramWL%blb(1:n), paramWL%bub(1:n), &
                    0.1_wp, tol, monitorQA, maxIter, Qform, ncall, iuser, ruser, ifail)

        ! Copy back from b vector and export parameters info
        paramWL%val(1:n) = paramWL%b(1:n) * paramWL%bscale(1:n) + paramWL%bmean(1:n)
        call paramSaveWL(paramWL,paramPath//'endParamWL.txt','Estimated parameters (Qform=' // realToStr(Qform) // ')')

        ! Check how far (final) simulated moments are from data moments
        if (ifail==0) print *, "Routine converged"
        call SolveAndSimWL(0)
        call SimMomentsWL(.true., momPath//'SimMomWL.txt', Qform, mom)
        write(*,'(/,"    Final Sum of Squared Differences in Moments: ", F15.6, /)') Qform
                
    end subroutine estimateWL


    !
    ! Calculate value of objective function at calling point in parameter space
    !
    subroutine objectiveCalcQA(n, b, Qform, iuser, ruser, inform)
        use functions, only : realToStr
        implicit none
        integer, intent(in)     :: n
        real(wp), intent(in)    :: b(n)
        real(wp), intent(out)   :: Qform
        integer, intent(inout)  :: iuser(*)
        real(wp), intent(inout) :: ruser(*)
        integer,intent(out)     :: inform
        integer  i
        real(wp) iQform

        iQform = Qform
        paramWL%val(1:n) = b(1:n) * paramWL%bscale(1:n) + paramWL%bmean(1:n)
        
        call getQ_WL(Qform)
        write(*,'(I4, A40, 2F12.6)') ncallQA, " Objective value at entrance and exit: ", iQform, Qform
        ncallQA = ncallQA+1
        
        inform = 1  !this needs to be +ve for the estimation to continue

    end subroutine objectiveCalcQA


    !
    ! Wrapper: Solves, simulates and calculates moments
    !
    subroutine getQ_WL(Qform)
        implicit none
        real(wp),intent(out) :: Qform
        type(momWL_t)        :: mom

        call SolveAndSimWL(0)
        call SimMomentsWL(.false., ' ', Qform, mom)
        Qform = Qform / (10.d0**QformScale)

    end subroutine getQ_WL



    !-----------------------------------------------------------------------------------------------------------------
    ! Routines to calculate the variance of estimated parameters
    !
        
    subroutine calcVarianceWL(npar,V)  
        implicit none
        integer,  intent(in)   :: npar
        real(wp), intent(out)  :: V(npar,npar)
        type(momWL_t)          :: mom        
        real(wp)               :: Qform, A(npar,npar), iA(npar,npar)
        real(wp), allocatable  :: Cmx(:,:), iCmx(:,:), dmom(:,:)
        integer                :: nmom, i, j


        ! Solve problem, simulate WL decisions, calculate moments and import 
        call SolveAndSimWL(0)
        call SimMomentsWL(.false., ' ', Qform, mom)

        ! Invert var-cov matrix of data moments
        nmom = ns * ( ng * ( mom%lWdistS(1)%n + mom%lWdistC(1)%n + mom%lWdistC_smsf(1)%n + &
                             mom%lWres(1)%n + mom%Eregr(1)%n + mom%Edist(1)%n ) + mom%lWjointC%n - 3 )
        if (allocated(Cmx)) deallocate(Cmx);  allocate (Cmx(nmom,nmom))
        if (allocated(iCmx)) deallocate(iCmx);  allocate (iCmx(nmom,nmom))
        call readMx(nmom, nmom, Cmx, mompath//'VmomWL.dta', 'V', 1)
        call invMx (nmom, Cmx, iCmx)
        
        ! Calculate derivatives of moments with respect to parameters
        if (allocated(dmom)) deallocate(dmom);  allocate (dmom (nmom,npar))        
        call getDerivativesWL(1.0e-2_wp, npar, nmom, dMom)

        ! Calculate iA(npar,npar) = ( dmom' iCmx dmom )^(-1)
        A = matmul( matmul(transpose(dmom),iCmx) , dMom )
        call invMx (npar, A, iA)

        ! Calculate VarCov of parameter estimates and store
        V = (1.0_wp + 1.0_wp/real(nrepsWL,wp)) * iA
        j = 0
        do i=1,paramWL%n
            if (paramWL%toestimate(i)==.true.) then
                j=j+1
                paramWL%se(i) = sqrt(V(j,j))
            end if
        end do
        
        ! Export to text parameters and variances
        call PrintEstimates(paramPath//'estimatesWL.txt', paramWL)
        call writeMx(npar, npar, V, paramPath//'CovWL.dta', 'V', 1)
            
    end subroutine calcVarianceWL


    ! 
    ! Calculate derivatives of moments with respect to each of the estimating parameters at estimated values
    !
    subroutine getDerivativesWL(delta, npar, nmom, dmom)
        implicit none
        real(wp), intent(in)    :: delta
        integer,  intent(in)    :: npar, nmom
        real(wp), intent(out)   :: dmom(nmom,npar)
        type(momWL_t)           :: mom
        real(wp)                :: estP, change, Qform(2), newmom(2,nmom), diff
        integer                 :: i, j, iter, n, m


        dmom = 0.0_wp
        n    = 0
        
        ! estimate derivative of each moment with respect to each parameter

        write (*,'(/,"  Calculating derivatives")')        
        do i = 1,paramWL%n
        if (paramWL%toEstimate(i) == .true.) then

            n    = n+1                              ! index for parameter in list of parameters being estimated
            estP = paramWL%val(i)                   ! keep original value of parameter

            do iter=0,3
            
                change = delta*(10.0_wp**real(iter,wp))
                
                ! compute moments and Qform after perturbing one parameter up or down
                do j = 1,2
                    if (j==1) paramWL%val(i) = max( estP-change, paramWL%lb(i) )
                    if (j==2) paramWL%val(i) = min( estP+change, paramWL%ub(i) )
                    call SolveAndSimWL(0)
                    call SimMomentsWL(.false., ' ', Qform(j), mom)
                    call storeMom(mom, nmom, newmom(j,:))
                end do

                ! calculate difference in Qform and check that it is non-zero
                diff =  min(estP+change, paramWL%ub(i)) - max(estP-change, paramWL%lb(i))
                write(*,'("  ",I3," ",A15,I2,F7.3,3F10.2)') i, paramWL%name(i), iter+1, diff, Qform(1), Qform(2), abs(Qform(1)-Qform(2))
                if (abs(Qform(1)-Qform(2)) > tol) exit
            
            end do
            
            ! approximate derivative of each moment with respect to parameter i
            do m=1,nmom
                dmom(m,n) = ( newmom(2,m) - newmom(1,m) ) / diff
            end do
            
            ! retrieve original value of parameter
            paramWL%val(i) = estP
            
        end if
        end do
        
    end subroutine getDerivativesWL


    !
    ! Store moments in same order as in VarCov matrix of moments estimates
    !
    subroutine storeMom(mom,totMom,Mvec)
        implicit none
        type(momWL_t), intent(in) :: mom
        integer, intent(in)       :: totMom
        real(wp),intent(out)      :: Mvec(totMom)
        integer                   :: m, s, j, g, ms, ws

        m = 0
        do g=1,ng
        do s=1,ns
            do j=1,mom%lWdistS(g)%n
                m=m+1;  Mvec(m)=mom%lWdistS(g)%sim(j,s)
            end do
            do j=mom%Eregr(g)%n/2+1,mom%Eregr(g)%n
                m=m+1;  Mvec(m)=mom%Eregr(g)%sim(j,s)
            end do
            do j=mom%Edist(g)%n/2+1,mom%Edist(g)%n
                m=m+1;  Mvec(m)=mom%Edist(g)%sim(j,s)
            end do
        end do
        end do

        do g=1,ng
        do s=1,ns
            do j=1,mom%lWdistC(g)%n
                m=m+1;  Mvec(m)=mom%lWdistC(g)%sim(j,s)
            end do
            do j=1,mom%Eregr(g)%n/2
                m=m+1;  Mvec(m)=mom%Eregr(g)%sim(j,s)
            end do
            do j=1,mom%Edist(g)%n/2
                m=m+1;  Mvec(m)=mom%Edist(g)%sim(j,s)
            end do
        end do
        end do

        do g=1,ng
        do s=1,ns
        do j=1,mom%lWres(g)%n
            m=m+1;  Mvec(m)=mom%lWres(g)%sim(j,s)
        end do
        end do
        end do

        do ms=1,3
        do ws=1,3
            m=m+1;  Mvec(m)=mom%lWdistC_smsf(1)%sim(ws,ms)          ! average earnings in couples by sm x sw and gender
            m=m+1;  Mvec(m)=mom%lWdistC_smsf(2)%sim(ms,ws)
            m=m+1;  Mvec(m)=mom%lWdistC_smsf(1)%sim(ws+ns,ms)       ! variance earnings in couples by sm x sw and gender
            m=m+1;  Mvec(m)=mom%lWdistC_smsf(2)%sim(ms+ns,ws)
            do j=1,mom%lWjointC%n/ns-1                              ! joint distribution of earnings in couples by sm x sw
                m=m+1;  Mvec(m)=mom%lWjointC%sim((ms-1)*mom%lWjointC%n/ns+j,ws)
            end do
        end do
        end do

    end subroutine storeMom




    !-----------------------------------------------------------------------------------------------------------------
    ! Auxialliary routines
    !

    !
    ! Trace optimisation procedure
    !
    subroutine monitorQA(n, ncall, b, Qform, rho, iuser, ruser, inform)
        implicit none
        integer,  intent(in)    :: n, ncall
        real(wp), intent(in)    :: b(n), Qform, rho
        integer,  intent(inout) :: iuser(1)
        real(wp), intent(inout) :: ruser(1)
        integer,  intent(out)   :: inform

        write (*,'(I5,A30,F15.7)') ncall, ' - current Qform value = ', Qform
        inform = 1  !this needs to be +ve for the estimation to continue

    end subroutine monitorQA


    !
    ! Rescale parameters for QA estimation to lie in (-1,1)
    !
    subroutine QAparameters(par)
        implicit none
        type(param_t), intent(inout) :: par
        integer                      :: i

        par%b       = 0.0_wp
        par%blb     = 0.0_wp
        par%bub     = 0.0_wp
        par%bmean   = 0.0_wp
        par%bscale  = 0.0_wp
        do i=1,par%n
            par%bmean(i)   = ( par%ub(i)+par%lb(i) ) * 0.5_wp
            par%bscale(i)  = par%ub(i)-par%bmean(i)
            par%b(i)       = merge( (par%val(i)-par%bmean(i))/par%bscale(i), par%b(i), abs(par%bscale(i))>tol )
            par%blb(i)     = merge( -1.0_wp, par%b(i), par%toEstimate(i)==.true. .and. abs(par%bscale(i))>tol )
            par%bub(i)     = merge(  1.0_wp, par%b(i), par%toEstimate(i)==.true. .and. abs(par%bscale(i))>tol )
        end do

    end subroutine QAparameters

end module estimationWL
