! --------------------------------------------------------------------------------------------------------------------
! This file is part of the package EMar, developed to produce the results in the paper "The marriage market, labor 
!    supply and education choice," by P.A. Chiapppori, M. Costa-Dias and C. Meghir. 
! Copyright (c) 2017 Monica Costa Dias
!
! Routines to simulate WL decisions (after education and marriage), for both couples and singles
! Conditional on education and marital choice as observed in data
! --------------------------------------------------------------------------------------------------------------------


module simulationWL

use precision,    only : wp, neginfty, tol
use paths,        only : dtapath, resultsPath
use constants,    only : mina, maxa, ng, ns, nH
use parameters,   only : paramCopyWL, paramWL, R, lnYC, lnyS, pQ, i2pQ, QminC, CminC, QminS, CminS, pthC, pthS, peta, &
                         kappa, SExi, SEu, SEe
use grids,        only : gKC, gKS, ngpk, gE, ngpe
use SimStructure, only : simLC, nobsLC, nrepsWL
use solutionWL,   only : iEiQ, mtxI, solveEE, getVdU, solveWL
use functions,    only : getH, earnings, logWres, LinInterp1D, LinInterp2D, resourcesC, utC, resourcesS, utS, &
                         getPosition
use IOroutines,   only : writeSimStata

real(wp),pointer :: pkC(:), peC(:,:), pwC(:,:), puC(:,:), pincC(:), pVC(:), pQC(:), plWresC(:,:), &
                    palphaC(:,:), pkS(:), peS(:), pwS(:), puS(:), pincS(:), pVS(:), pQS(:), plWresS(:), palphaS(:)
integer, pointer :: pLC(:,:), pLS(:)
real(wp), private, allocatable :: rve1(:,:,:), rvxi(:,:,:,:), rvu(:,:,:,:), rvtheta(:,:), rveta(:,:,:)


contains


    !-----------------------------------------------------------------------------------------------------------------
    ! Wrapper: solution and simulation of LC problem
    !
    
    subroutine SolveAndSimWL(dwrite)
        implicit none
        integer, intent(in) :: dwrite

        call paramCopyWL(paramWL)
        call SolveWL
        call simulateWL(dwrite)    
            
    end subroutine SolveAndSimWL



    !-----------------------------------------------------------------------------------------------------------------
    ! Simulate post-marital choices of each individual/couple
    ! These simulations are conditional on the education and marital status observed in data, so
    !    are only for those observed during working life (insampleWL==1)
    !

    subroutine simulateWL(dwrite)
        use nag_g_ib,  only : g05kbf
        implicit none
        integer, intent(in) :: dwrite
        integer             :: rep, i, ietam, ietaf, ieta, ithm, ithf, ith, sm, sf, s, g, a
        real(wp)            :: discY(ng,maxa-mina+1)


        ! Read initial conditions and draw shocksaffecting choices after marriage: 
        ! standardised (e1, xi, u) in logs and (eta, theta)
        call ReadInitWL
        call DrawShocks


        ! simulate choice for couples, mina-maxa
        do i=1,nobsLC
        
            if ( simLC(i)%insampleWL==1 .and. simLC(i)%coupleDta==1) then

                do rep=1,nrepsWL
                    call ilocateC(i,rep,sm,sf,ithm,ithf,ietam,ietaf)       ! locate matrices and fill info for shocks
                    call simulateC(sm,sf,ithm,ithf,ietam,ietaf)
                    discY = 0.0_wp
                    do g=1,ng
                        forall (a=1:maxa-mina+1) discY(g,a) = (R**real(1-a,wp)) * pwC(g,a) * real(1-pLC(g,a),wp)
                        simLC(i)%wl(1,rep)%discY(g) = sum(discY(g,:))
                    end do
                end do

            else if ( simLC(i)%insampleWL==1 .and. simLC(i)%coupleDta==0 ) then

                g = simLC(i)%female+1
                do rep=1,nrepsWL            
                    call ilocateS(i,rep,g,s,ith,ieta)       ! locate matrices and fill info for shocks
                    call simulateS(g,s,ith,ieta)
                    discY = 0.0_wp
                    forall (a=1:maxa-mina+1) discY(g,a) = (R**real(1-a,wp)) * pwS(a) * real(1-pLS(a),wp)
                    simLC(i)%wl(1,rep)%discY(g) = sum(discY(g,:))
                end do
            
            end if
        end do

        if (dwrite==1) call writesimstata(resultsPath//"simWL.dta")

    end subroutine simulateWL


    !
    ! Simulate post-marital decisions of a couple
    !
    subroutine simulateC(sm,sf,ithm,ithf,ietam,ietaf)
        implicit none
        integer,intent(in) :: sm,sf,ithm,ithf,ietam,ietaf
        integer            :: a, ia, ik, L(ng,4), Lm, Lf, iL, maxL(1), ubem, ubef, ihm, ihf
        real(wp)           :: k_1, wm, wf, V(4), iQ(4), K(4), Kmin, Kmax, resK0, I1(ngpk), iEiQ1(ngpk), alpha(ng,4)
        real(wp), pointer  :: nxtI(:,:,:), nxtiEiQ(:,:,:)
        

        ! calculate human capital
        ihm = getH(ithm,sm)
        ihf = getH(ithf,sf)

        ! solve problem of couple i
        do a = mina,maxa

            ia   = a-mina+1       ! age index for pointer
            k_1  = 0.0_wp;   if (a>mina) k_1 = pkC(ia-1)
            wm   = earnings(1, sm, lnyC(1,ithm,sm), a, peC(1,ia))
            wf   = earnings(2, sf, lnyC(2,ithf,sf), a, peC(2,ia))
            V    = neginfty
            iQ   = 0.0_wp
            L    = 0
            K    = 0.0_wp
            Kmin = gKC(1,sm,sf,ithm,ithf,a)
            
            ! interpolate continuation value over productivity
            if (a<maxa) then
                nxtI    => mtxI(a+1)%C(ietam,ietaf,ihm,ihf,:,:,:)
                nxtiEiQ => iEiQ(a+1)%C(ietam,ietaf,ihm,ihf,:,:,:)
                call getPosition(peC(1,ia), ngpe, ge(1,sm,:), ubem, 1)
                call getPosition(peC(2,ia), ngpe, ge(2,sf,:), ubef, 1)
                do ik=1,ngpk
                    I1(ik)    = LinInterp2D ( peC(1,ia), ge(1,sm,ubem-1), ge(1,sm,ubem), peC(2,ia), ge(2,sf,ubef-1), ge(2,sf,ubef), &
                                              nxtI(ik,ubem-1,ubef-1), nxtI(ik,ubem-1,ubef), nxtI(ik,ubem,ubef-1), nxtI(ik,ubem,ubef) )
                    iEiQ1(ik) = LinInterp2D ( peC(1,ia), ge(1,sm,ubem-1), ge(1,sm,ubem), peC(2,ia), ge(2,sf,ubef-1), ge(2,sf,ubef), &
                                              nxtiEiQ(ik,ubem-1,ubef-1), nxtiEiQ(ik,ubem-1,ubef), nxtiEiQ(ik,ubem,ubef-1), nxtiEiQ(ik,ubem,ubef) )
                end do
            end if

            ! solve intertemporal condition for each LS choice                
            do Lm = 0,1
            do Lf = 0,1
                
                iL        = Lm*2+Lf+1
                L(1,iL)   = Lm
                L(2,iL)   = Lf
                resK0     = resourcesC (wm, wf, Lm, Lf, k_1, 0.0_wp)
                Kmax      = min ( resK0-pQ*QminC-CminC , gKC(ngpk,sm,sf,ithm,ithf,a) )
                alpha(1,iL) = real(Lm,wp)*utC(1,ietam,sm,a,puC(1,ia))
                alpha(2,iL) = real(Lf,wp)*utC(2,ietaf,sf,a,puC(2,ia))

                if (a<maxa) then
                    call solveEE(gKC(:,sm,sf,ithm,ithf,a), alpha(1,iL)+alpha(2,iL), Kmin, Kmax, resK0, iEiQ1, I1, K(iL), V(iL), iQ(iL))
                else if (Kmax-Kmin>-tol) then
                    K(iL) = 0.0_wp
                    call getVdU(resK0, K(iL), 0.0_wp, 0.0_wp, 0.0_wp, 0.0_wp, alpha(1,iL)+alpha(2,iL), V(iL), iQ(iL))
                end if                
                
            end do
            end do

            ! now select optimum and store it
            maxL          = maxloc(V)
            pkC(ia)       = K(maxL(1))
            pLC(1,ia)     = L(1,maxL(1))
            pLC(2,ia)     = L(2,maxL(1))
            palphaC(1,ia) = alpha(1,maxL(1))
            palphaC(2,ia) = alpha(2,maxL(1))
            pwC(1,ia)     = wm
            pwC(2,ia)     = wf
            plWresC(1,ia) = logWres(pwC(1,ia),1,sm,a)
            plWresC(2,ia) = logWres(pwC(2,ia),2,sf,a)
            pincC(ia)     = resourcesC (wm, wf, pLC(1,ia), pLC(2,ia), k_1, 0.0_wp)
            pVC(ia)       = V(maxL(1))
            pQC(ia)       = ( pincC(ia) - pkC(ia) + palphaC(1,ia) + palphaC(2,ia) ) * i2pQ
            
        end do
            
    end subroutine simulateC


    !
    ! Simulate post-marital decisions of a single individual
    !
    subroutine simulateS(g,s,ith,ieta)
        implicit none
        integer,intent(in) :: g,s,ith,ieta
        integer            :: a, ia, ik, L, maxL(1), ube, ih
        real(wp)           :: k_1, w, V(0:1), iQ(0:1), K(0:1), Kmin, Kmax, resK0, I1(ngpk), iEiQ1(ngpk), alpha(0:1)
        real(wp), pointer  :: nxtI(:,:), nxtiEiQ(:,:)
        

        ! calculate human capital
        ih = getH(ith,s)

        ! solve problem of couple i
        do a = mina,maxa

            ia   = a-mina+1       ! age index for pointer
            k_1  = 0.0_wp;   if (a>mina) k_1 = pkS(ia-1)
            w    = earnings(g, s, lnyS(g,ith,s), a, peS(ia))
            V    = neginfty
            iQ   = 0.0_wp
            L    = 0
            K    = 0.0_wp
            Kmin = gKS(1,s,ith,a,g)
            
            ! interpolate continuation value over productivity
            if (a<maxa) then
                nxtI    => mtxI(a+1)%S(g,ieta,ih,:,:)
                nxtiEiQ => iEiQ(a+1)%S(g,ieta,ih,:,:)
                call getPosition(peS(ia), ngpe, ge(g,s,:), ube, 1)
                do ik=1,ngpk
                    I1(ik)    = LinInterp1D ( peS(ia), ge(g,s,ube-1), ge(g,s,ube), nxtI(ik,ube-1), nxtI(ik,ube) )
                    iEiQ1(ik) = LinInterp1D ( peS(ia), ge(g,s,ube-1), ge(g,s,ube), nxtiEiQ(ik,ube-1), nxtiEiQ(ik,ube) )
                end do
            end if

            ! solve intertemporal condition for each LS choice                
            do L = 0,1
                
                resK0    = resourcesS(w,L,k_1,0.0_wp)
                Kmax     = min ( resK0-pQ*QminS-CminS , gKS(ngpk,s,ith,a,g) )
                alpha(L) = real(L,wp)*utS(g,ieta,s,a,puS(ia))
                
                if (a<maxa) then
                    call solveEE(gKS(:,s,ith,a,g), alpha(L), Kmin, Kmax, resK0, iEiQ1, I1, K(L), V(L), iQ(L))
                else if (Kmax-Kmin>-tol) then
                    K(L) = 0.0_wp
                    call getVdU(resK0, K(L), 0.0_wp, 0.0_wp, 0.0_wp, 0.0_wp, alpha(L), V(L), iQ(L))
                end if                
                
            end do

            ! now select optimum and store it
            maxL        = maxloc(V)
            pkS(ia)     = K(maxL(1)-1)
            pLS(ia)     = maxL(1)-1
            palphaS(ia) = alpha(maxL(1)-1)
            pwS(ia)     = w
            plWresS(ia) = logWres(pwS(ia),g,s,a)
            pincS(ia)   = resourcesS (w, pLS(ia), k_1, 0.0_wp)
            pVS(ia)     = V(maxL(1)-1)
            pQS(ia)     = ( pincS(ia) - K(maxL(1)-1) + alpha(maxL(1)-1) ) * i2pQ
            
        end do
            
    end subroutine simulateS


    !
    ! Initialise simulation for couple i
    !
    subroutine ilocateC(i,rep,sm,sf,ithm,ithf,ietam,ietaf)
        implicit none
        integer,intent(in)  :: i,rep
        integer,intent(out) :: sm,sf,ithm,ithf,ietam,ietaf
        integer             :: a, ia


        ! get education from data
        sm = simLC(i)%s(1)
        sf = simLC(i)%s(2)

        ! get unobserved heterogeneity: ability
        ithm=2; ithf=2
        if ( rvtheta(i,rep)<pthC(1,1,sm,sf)) then
            ithm=1; ithf=1
        else if ( rvtheta(i,rep)>=pthC(1,1,sm,sf) .and. rvtheta(i,rep)<pthC(1,1,sm,sf)+pthC(1,2,sm,sf) ) then
            ithm=1
        else if ( rvtheta(i,rep)>=pthC(1,1,sm,sf)+pthC(1,2,sm,sf) .and. rvtheta(i,rep)<pthC(1,1,sm,sf)+pthC(1,2,sm,sf)+pthC(2,1,sm,sf) ) then
            ithf=1
        end if
        simLC(i)%wl(1,rep)%itheta(1)=ithm
        simLC(i)%wl(1,rep)%itheta(2)=ithf
            
        ! get unobserved heterogeneity: preferences
        ietam=1;   if (rveta(1,i,rep)>=peta(1,sm)) ietam=2;    simLC(i)%wl(1,rep)%ieta(1)=ietam
        ietaf=1;   if (rveta(2,i,rep)>=peta(2,sf)) ietaf=2;    simLC(i)%wl(1,rep)%ieta(2)=ietaf
                        
        ! locate area in matrices to fill
        pKC     => simLC(i)%wl(1,rep)%k(:)
        peC     => simLC(i)%wl(1,rep)%e(:,:)
        pwC     => simLC(i)%wl(1,rep)%w(:,:)
        plWresC => simLC(i)%wl(1,rep)%lWres(:,:)
        puC     => simLC(i)%wl(1,rep)%u(:,:)
        pincC   => simLC(i)%wl(1,rep)%inc(:)
        pVC     => simLC(i)%wl(1,rep)%V(:)
        pQC     => simLC(i)%wl(1,rep)%Q(:)
        pLC     => simLC(i)%wl(1,rep)%L(:,:)
        palphaC => simLC(i)%wl(1,rep)%alpha(:,:)

        ! calculate productivity
        peC(1,1) = rve1(1,i,rep)*SEe(1,sm)
        peC(2,1) = rve1(2,i,rep)*SEe(2,sf)
        do a=mina+1,maxa
            ia = a-mina+1
            peC(1,ia) = peC(1,ia-1)*kappa(1,sm) + rvxi(1,i,rep,a)*SExi(1,sm)
            peC(2,ia) = peC(2,ia-1)*kappa(2,sf) + rvxi(2,i,rep,a)*SExi(2,sf)
        end do

        ! and transitory preferences for leisure
        do a=mina,maxa
            ia = a-mina+1
            puC(1,ia) = rvu(1,i,rep,a)*SEu(1,sm)
            puC(2,ia) = rvu(2,i,rep,a)*SEu(2,sf)
        end do

    end subroutine ilocateC


    !
    ! Initialise simulation for single individual i
    !
    subroutine ilocateS(i,rep,g,s,ith,ieta)
        implicit none
        integer,intent(in)  :: i,rep,g
        integer,intent(out) :: ith,ieta,s
        integer             :: a, ia


        ! get education
        s = simLC(i)%s(g)

        ! get unobserved heterogeneity
        ith=1;     if (rvtheta(i,rep)>=pthS(g,ith,s)) ith=2;    simLC(i)%wl(1,rep)%itheta(g)=ith
        ieta=1;    if (rveta(g,i,rep)>=peta(g,s)) ieta=2;       simLC(i)%wl(1,rep)%ieta(g)=ieta

        ! locate area in matrices to fill
        pKS     => simLC(i)%wl(1,rep)%k(:)
        peS     => simLC(i)%wl(1,rep)%e(g,:)
        pwS     => simLC(i)%wl(1,rep)%w(g,:)
        plWresS => simLC(i)%wl(1,rep)%lWres(g,:)
        puS     => simLC(i)%wl(1,rep)%u(g,:)
        pincS   => simLC(i)%wl(1,rep)%inc(:)
        pVS     => simLC(i)%wl(1,rep)%V(:)
        pQS     => simLC(i)%wl(1,rep)%Q(:)
        pLS     => simLC(i)%wl(1,rep)%L(g,:)
        palphaS => simLC(i)%wl(1,rep)%alpha(g,:)

        ! calculate productivity
        peS(1) = rve1(g,i,rep)*SEe(g,s)
        do a=mina+1,maxa
            ia = a-mina+1
            peS(ia) = peS(ia-1)*kappa(g,s) + rvxi(g,i,rep,a)*SExi(g,s)
        end do

        ! and preferences for leisure
        do a=mina,maxa
            ia = a-mina+1
            puS(ia) = rvu(g,i,rep,a)*SEu(g,s)
        end do

    end subroutine ilocateS



    !-----------------------------------------------------------------------------------------------------------------
    ! Read data: education, marital status and observation period
    ! Than clean and fill simulation structure
    !
    
    subroutine ReadInitWL
        use stataMod
        implicit none
        integer :: unit, rep, i


        ! read BHPS data
        unit = 5454
        call openStata(dtaPath//'BHPSsample.dta', unit)
        nobsLC = nobsStata()
        if (allocated(simLC)) deallocate(simLC);   allocate (simLC(nobsLC))
        call readStata(simLC(:)%coupleDta,  'couple'       )
        call readStata(simLC(:)%female,     'w'            )
        call readStata(simLC(:)%insampleWL, 'insampleWL'   )
        call readStata(simLC(:)%insampleS,  'insampleEduc' )
        call readStata(simLC(:)%age1,       'age1'         )
        call readStata(simLC(:)%ageN,       'ageN'         )
        call readStata(simLC(:)%s(1),       'mSadult'      )
        call readStata(simLC(:)%s(2),       'wSadult'      )
        call closeOpenStata() 

        ! clean and fill simulation structure
        do i=1,nobsLC
        do rep=1,nrepsWL
            simLC(i)%wl(1,rep)%k        = 0.0_wp
            simLC(i)%wl(1,rep)%e        = 0.0_wp
            simLC(i)%wl(1,rep)%w        = 0.0_wp
            simLC(i)%wl(1,rep)%lWres    = 0.0_wp
            simLC(i)%wl(1,rep)%u        = 0.0_wp
            simLC(i)%wl(1,rep)%inc      = 0.0_wp
            simLC(i)%wl(1,rep)%V        = 0.0_wp
            simLC(i)%wl(1,rep)%L        = 0
            simLC(i)%wl(1,rep)%alpha    = 0.0_wp
            simLC(i)%wl(1,rep)%Q        = 0.0_wp
            simLC(i)%wl(1,rep)%ieta     = 0
            simLC(i)%wl(1,rep)%itheta   = 0
            simLC(i)%wl(1,rep)%insample = 0
            if (simLC(i)%insampleWL==1) simLC(i)%wl(1,rep)%insample( max(mina,simLC(i)%age1) : min(50,simLC(i)%ageN) ) = 1.0_wp
        end do
        end do
 
    end subroutine ReadInitWL
    


    !-----------------------------------------------------------------------------------------------------------------
    ! Generate random draws for shocks afftecting choices after marriage: standard deviates only
    !
    
    subroutine DrawShocks
        use nag_g_ib,  only : g05laf, g05kbf, g05lgf
        implicit none
        integer ifail, igen, iseed(4)
        real(wp) mu, se


        ! initialisations
        if (allocated(rvtheta)) deallocate(rvtheta);  allocate (rvtheta(nobsLC,nrepsWL))
        if (allocated(rveta))   deallocate(rveta);    allocate (rveta  (ng,nobsLC,nrepsWL))
        if (allocated(rve1))    deallocate(rve1);     allocate (rve1   (ng,nobsLC,nrepsWL))
        if (allocated(rvxi))    deallocate(rvxi);     allocate (rvxi   (ng,nobsLC,nrepsWL,mina:maxa))
        if (allocated(rvu))     deallocate(rvu);      allocate (rvu    (ng,nobsLC,nrepsWL,mina:maxa))

        mu = 0.0_wp
        se = 1.0_wp

        igen     = 0
        iseed    = 0
        iseed(1) = 8504165
        call g05kbf(igen,iseed)
        
        ! Draw normal deviates randomly for productivity initial productivity in logs, standardised (e1)
        igen  = 0
        ifail = 0
        call g05laf (mu, se**2.0_wp, ng*nobsLC*nrepsWL, rve1(:,:,:), igen, iseed, ifail)

        ! Draw normal deviates randomly for innov in productivity in logs, standardised (xi)
        igen  = 0
        ifail = 0
        call g05laf (mu, se**2.0_wp, ng*nobsLC*nrepsWL*(maxa-mina+1), rvxi(:,:,:,:), igen, iseed, ifail)

        ! Draw normal deviates randomly for transitory preferences for leisure in logs, standardised (u)
        igen  = 0
        ifail = 0
        call g05laf (mu, se**2.0_wp, ng*nobsLC*nrepsWL*(maxa-mina+1), rvu(:,:,:,:), igen, iseed, ifail)

        ! Draw uniform deviates randomly for ability (productivity) theta
        igen  = 0
        ifail = 0
        call g05lgf(0.0_wp, 1.0_wp, nobsLC*nrepsWL, rvtheta(:,:), igen, iseed, ifail)

        ! Draw uniform deviates randomly for preferences for working, eta
        igen  = 0
        ifail = 0
        call g05lgf(0.0_wp, 1.0_wp, ng*nobsLC*nrepsWL, rveta(:,:,:), igen, iseed, ifail)

    end subroutine DrawShocks
    
end module simulationWL


