! --------------------------------------------------------------------------------------------------------------------
! This file is part of the package EMar, developed to produce the results in the paper "The marriage market, labor 
!    supply and education choice," by P.A. Chiapppori, M. Costa-Dias and C. Meghir. 
! Copyright (c) 2017 Monica Costa Dias
!
! Routines to solve dynamic problem after education and marital choics, for couples and singles
! --------------------------------------------------------------------------------------------------------------------


module solutionWL

use precision,  only : wp, neginfty, tol
use constants,  only : ng, ns, neta, nH, mina, maxa
use parameters, only : ntheta, lnYC, lnYS, QminC, QminS, CminC, CminS, pQ, i2pQ, lnpQ, iBR, beta, derived_parWL
use grids,      only : ngpk, ngpe, ngpu, gKC, gKS, gE, gU, TMe
use functions,  only : earnings, resourcesC, resourcesS, utC, utS, getH, LinInterp1D, LinInterp2D, getGridShocks, getGridK
use omp_lib

type mtxE
    real(wp) :: C(neta,neta,nH,nH,ngpk,ngpe,ngpe)   ! expected values for couples; dimensions: (meta, weta, mh, wh, k(a-1), me(a-1), we(a-1))
    real(wp) :: S(ng,neta,nH,ngpk,ngpe)             ! expected values for singles; dimensions: (g, eta, h, k(a-1), e(a-1))
end type mtxE
type(mtxE), target :: iEiQ(mina:maxa)               ! storage matrix for argument in Euler Equation
type(mtxE), target :: mtxI(mina:maxa)               ! storage matrix for expected continuation value excl Pareto weight

real(wp), parameter, private :: wg_uC = 1.0_wp / real(ngpu*ngpu,wp), &
                                wg_uS = 1.0_wp / real(ngpu,wp)

                                
contains


    ! ----------------------------------------------------------------------------------------------------------------
    ! Wrapper: solution of WL problem, by marital status and education
    !
    subroutine SolveWL
        implicit none

        ! Prepare current version of coeffs and grids
        call derived_parWL
        call getGridShocks
        call getGridK
    
        ! Solve individual's problem and simulate LC profiles
        call solveC
        call solveS
            
    end subroutine SolveWL




    ! ----------------------------------------------------------------------------------------------------------------
    ! Couples: solution wrapper and solver routines by age
    !
    
    !
    ! Couples: wrapper
    !
    subroutine solveC
        implicit none
        integer  a, sm, sf, ithm, ithf, ik_1, ietam, ietaf, ihm, ihf


        ! age maxa
        call slnC_maxa
        
        ! all earlier ages
        do a = maxa-1,mina,-1
            do ietam = 1,neta
            do ietaf = 1,neta
            do ithm  = 1,ntheta
            do ithf  = 1,ntheta
            do sm    = 1,ns
            do sf    = 1,ns
                !$OMP PARALLEL DO DEFAULT(NONE) &
                !$OMP SHARED(a, ietam, ietaf, sm, sf, ithm, ithf, iEiQ, mtxI) &  !! EVm, surplus) &
                !$OMP PRIVATE(ik_1)
                do ik_1 = 1,ngpk
                    call slnC_a(a, ietam, ietaf, sm, sf, ithm, ithf, ik_1)
                end do
                !$OMP END PARALLEL DO
            end do
            end do
            end do
            end do
            end do
            end do
        end do

    end subroutine solveC


    !
    ! Couples: solution for age=mina,...,maxa-1
    ! solves problem at (a,etam,etaf,hm(sm,thm),hf(sf,thf),k_1) for all (em,ef,um,uf) and calculates
    !     expected continuation value and marginal utility conditional on information in t-1, (e_1,a_1)
    !
    subroutine slnC_a(a, ietam, ietaf, sm, sf, ithm, ithf, ik_1)
        implicit none
        integer, intent(in) :: a,  ietam, ietaf, sm, sf, ithm, ithf, ik_1
        integer  :: i, ihm, ihf, iem, ief, iem_1, ief_1, ium, iuf, Lm, Lf, num, nuf, maxL(1)
        real(wp) :: k_1, resK0, wm, wf, Kmin, Kmax, minExp, k, I1k, alpha, &
                    V(4,ngpu,ngpu), iQ(4,ngpu,ngpu), optV(ngpu,ngpu), optiQ(ngpu,ngpu), &
                    EV(ngpe,ngpe), EiQ(ngpe,ngpe), intEV(ngpe), intEiQ(ngpe)


        ! Initialise matrices and calculate quantities needed in solution
        EV     = neginfty
        intEV  = neginfty
        EiQ    = 0.0_wp
        intEiQ = 0.0_wp
        k_1    = gKC(ik_1,sm,sf,ithm,ithf,a-1)
        Kmin   = gKC(1,sm,sf,ithm,ithf,a)
        ihm    = getH(ithm,sm)
        ihf    = getH(ithf,sf)

        
        ! Solve problem conditional on productivity shocks at time a, for each type in preferences (eta)
        do iem = 1,ngpe
        do ief = 1,ngpe

            ! calculate earnings rate
            wm = earnings(1, sm, lnyC(1,ithm,sm), a, ge(1,sm,iem))
            wf = earnings(2, sf, lnyC(2,ithf,sf), a, ge(2,sf,ief))            
            
            ! calculate optimal savings and their value by labour supply of spouses
            do Lm=0,1
            do Lf=0,1

                i     = Lm*2+Lf+1
                num   = merge(1,ngpu,Lm==0)
                nuf   = merge(1,ngpu,Lf==0)
                resK0 = resourcesC (wm, wf, Lm, Lf, k_1, 0.0_wp)
                Kmax  = min ( resK0-pQ*QminC-CminC , gKC(ngpk,sm,sf,ithm,ithf,a) )
                        
                ! calculate solution to euler equation
                do ium=1,num
                do iuf=1,nuf
                    alpha = real(Lm,wp)*utC(1,ietam,sm,a,gu(1,sm,ium)) + real(Lf,wp)*utC(2,ietaf,sf,a,gu(2,sf,iuf))
                    call solveEE(gKC(:,sm,sf,ithm,ithf,a), alpha, Kmin, Kmax, resK0, &
                                 iEiQ(a+1)%C(ietam,ietaf,ihm,ihf,:,iem,ief), mtxI(a+1)%C(ietam,ietaf,ihm,ihf,:,iem,ief), &
                                 k, V(i,ium,iuf), iQ(i,ium,iuf))
                end do
                end do
        
                ! fill in missing gaps: irrelevant combinations of preference shocks (uf,um) when at least
                !   one spouse works
                do ium=1,num
                do iuf=nuf+1,ngpu
                    V(i,ium,iuf)  = V(i,ium,nuf)
                    iQ(i,ium,iuf) = iQ(i,ium,nuf)
                end do
                end do
                do ium=num+1,ngpu
                    V(i,ium,:)  = V(i,num,:)
                    iQ(i,ium,:) = iQ(i,num,:)
                end do
                
            end do
            end do
            
            ! determine optimal labour supply and store value
            optV  = 0.0_wp
            optiQ = 0.0_wp
            do ium=1,ngpu
            do iuf=1,ngpu
                maxL           = maxloc(V(:,ium,iuf))
                optV(ium,iuf)  = V(maxL(1),ium,iuf)       ! = 2*ln(Q)+ln(pQ) + beta*I(t+1) at the optimal LS
                optiQ(ium,iuf) = iQ(maxL(1),ium,iuf)      ! = Q(t)^(-1) at the optimal LS
                if (optV(ium,iuf) < neginfty+tol) then    ! check for error
                    write(*,*) 'no possible choice of employment is affordable', a, sm, sf
                    stop
                end if
            end do
            end do
            
            ! integrate out preference shocks
            EV(iem, ief)  = sum(optV)*wg_uC
            EiQ(iem, ief) = sum(optiQ)*wg_uC

        end do
        end do


        ! Integrate out productivity conditional on productivity in previous period (a-1)
        do iem_1=1,ngpe
            forall (ief=1:ngpe) intEV(ief)  = dot_product( TMe(1,sm,a,iem_1,:), EV(:,ief)  ) 
            forall (ief=1:ngpe) intEiQ(ief) = dot_product( TMe(1,sm,a,iem_1,:), EiQ(:,ief) )
            do ief_1=1,ngpe
                mtxI(a)%C(ietam,ietaf,ihm,ihf,ik_1,iem_1,ief_1) = dot_product( TMe(2,sf,a,ief_1,:), intEV )
                iEiQ(a)%C(ietam,ietaf,ihm,ihf,ik_1,iem_1,ief_1) = 1.0_wp / dot_product( TMe(2,sf,a,ief_1,:), intEiQ )
            end do
        end do

    end subroutine slnC_a
    


    ! 
    ! Couples: solution at age maxa
    !
    subroutine slnC_maxa
        implicit none
        integer  :: a, ietam, ietaf, sm, sf, ihm, ihf, ithm, ithf, ik_1, i, Lm, Lf, maxL(1), iem, ief, iem_1, ief_1, ium, iuf
        real(wp) :: k_1, wm, wf, alpha, resK0, &
                    Q(4,ngpu,ngpu), optV(ngpu,ngpu), optiQ(ngpu,ngpu), EV(ngpe,ngpe), EiQ(ngpe,ngpe), intEV(ngpe), intEiQ(ngpe)


        ! Initialise storage matrices 
        forall (a=mina:maxa)
            iEiQ(a)%C = 0.0_wp
            mtxI(a)%C = 0.0_wp
        end forall

        
        ! Calculate continuation value at maxa conditional on savings at the end of period k(maxa)=0
        a = maxa
        do ietam = 1,neta
        do ietaf = 1,neta
        do ithm  = 1,ntheta
        do ithf  = 1,ntheta
        do sm    = 1,ns
        do sf    = 1,ns

            ihm = getH(ithm,sm)
            ihf = getH(ithf,sf)
        
            ! savings at maxa-1 ensure minQ is affordable in future (maxa)
            do ik_1 = 1,ngpk  

                k_1 = gkC(ik_1,sm,sf,ithm,ithf,a-1)

                do iem   = 1,ngpe
                do ief   = 1,ngpe

                    ! calculate earnings rate
                    wm = earnings(1, sm, lnyC(1,ithm,sm), a, ge(1,sm,iem))
                    wf = earnings(2, sf, lnyC(2,ithf,sf), a, ge(2,sf,ief))
                    
                    ! calculate value of each possible LS point
                    do Lm=0,1
                    do Lf=0,1
                        i     = Lm*2+Lf+1
                        resK0 = resourcesC (wm, wf, Lm, Lf, k_1, 0.0_wp)
                        do ium = 1,ngpu
                        do iuf = 1,ngpu
                            alpha        = real(Lm,wp)*utC(1,ietam,sm,a,gu(1,sm,ium)) + real(Lf,wp)*utC(2,ietaf,sf,a,gu(2,sf,iuf))
                            Q(i,ium,iuf) = (resK0 + alpha) * i2pQ
                        end do
                        end do
                    end do
                    end do

                    ! select optimal LS and integrate out taste shocks
                    do ium = 1,ngpu
                    do iuf = 1,ngpu
                        maxL = maxloc(Q(:,ium,iuf))
                        optV(ium,iuf)  = 2.0_wp * log(Q(maxL(1),ium,iuf)) + lnpQ
                        optiQ(ium,iuf) = 1.0_wp / Q(maxL(1),ium,iuf)
                    end do
                    end do
                    EV(iem,ief)  = sum(optV)*wg_uC
                    EiQ(iem,ief) = sum(optiQ)*wg_uC
                    
                end do !ef
                end do !em

                ! integrate out productivity conditional on productivity at a-1
                do iem_1=1,ngpe
                    forall (ief=1:ngpe) intEV(ief)  = dot_product( TMe(1,sm,a,iem_1,:), EV(:,ief)  ) 
                    forall (ief=1:ngpe) intEiQ(ief) = dot_product( TMe(1,sm,a,iem_1,:), EiQ(:,ief) )
                    do ief_1=1,ngpe
                        mtxI(a)%C(ietam,ietaf,ihm,ihf,ik_1,iem_1,ief_1) = dot_product( TMe(2,sf,a,ief_1,:), intEV )
                        iEiQ(a)%C(ietam,ietaf,ihm,ihf,ik_1,iem_1,ief_1) = 1.0_wp / dot_product( TMe(2,sf,a,ief_1,:), intEiQ )
                    end do
                end do
                
            end do !k_1               
            
        end do !thf
        end do !thm
        end do !sf
        end do !sm
        end do !etaf
        end do !etam

    end subroutine slnC_maxa




    ! ----------------------------------------------------------------------------------------------------------------
    ! Singles: solution wrapper and solver routines by age
    !
    
    !
    ! Singles: wrapper
    !
    subroutine solveS
        implicit none
        integer  g, a, s, ith, ik_1, ieta, ih

        do g=1,ng

            ! age maxa
            call slnS_maxa(g)
        
            ! ages mina+1 to maxa-1
            do a = maxa-1,mina,-1
                do ieta = 1,neta
                do ith  = 1,ntheta
                do s    = 1,ns
                    !$OMP PARALLEL DO DEFAULT(NONE) &
                    !$OMP SHARED(g, ieta, a, s, ith, iEiQ, mtxI) &
                    !$OMP PRIVATE(ik_1)
                    do ik_1 = 1,ngpk
                        call slnS_a(g, a, ieta, s, ith, ik_1)
                    end do
                    !$OMP END PARALLEL DO
                end do
                end do
                end do
            end do

        end do

    end subroutine solveS



    !
    ! Singles: solution at age=mina,...,maxa-1
    ! for each (g,a,eta,e,th) solves problem for all (e,u) and computes expected continuation 
    !    value and marginal utility conditional on info in t-1 (e_1,k_1)
    !
    subroutine slnS_a(g, a, ieta, s, ith, ik_1)
        implicit none
        integer, intent(in) :: g, a,  ieta, s, ith, ik_1
        integer  :: ih, L, ie, ie_1, iu, maxL(1), nu
        real(wp) :: w, resK0, alpha, k_1, Kmin, Kmax, k, V(0:1,ngpu), iQ(0:1,ngpu), optV(ngpu), optiQ(ngpu), EV(ngpe), EiQ(ngpe)


        ! Initialise matrices and calculate quantities needed in solution
        EV     = neginfty
        EiQ    = 0.0_wp
        k_1    = gKS(ik_1,s,ith,a-1,g)
        Kmin   = gKS(1,s,ith,a,g)
        ih     = getH(ith,s)


        ! Solve problem conditional on productivity shocks at time a, for each type in preferences eta
        do ie = 1,ngpe

            ! calculate earnings rate
            w = earnings(g, s, lnyS(g,ith,s), a, ge(g,s,ie))
            
            ! calculate optimal savings and its value
            do L=0,1

                nu    = merge(1,ngpu,L==0)
                resK0 = resourcesS (w, L, k_1, 0.0_wp)
                Kmax  = min ( resK0-pQ*QminS-CminS , gKS(ngpk,s,ith,a,g) )
                        
                ! calculate solution to EE
                do iu=1,nu
                    alpha = real(L,wp)*utS(g,ieta,s,a,gu(g,s,iu))
                    call solveEE(gKS(:,s,ith,a,g), alpha, Kmin, Kmax, resK0, &
                                 iEiQ(a+1)%S(g,ieta,ih,:,ie), mtxI(a+1)%S(g,ieta,ih,:,ie), k, V(L,iu), iQ(L,iu))
                end do
        
                ! now fill in missing gaps: irrelevant combinations of (uf,um) when some spouse is working
                do iu=nu+1,ngpu
                    V(L,iu)  = V(L,nu)
                    iQ(L,iu) = iQ(L,nu)
                end do
                
            end do
            
            ! determine optimal LS and store value
            optV   = 0.0_wp
            optiQ = 0.0_wp
            do iu=1,ngpu 
                maxL      = maxloc(V(:,iu))       ! optimal leisure (+1)
                optV(iu)  = V(maxL(1)-1,iu)       ! = 2*ln(Q)+ln(pQ) + beta*I(t+1) at the optimal LS point
                optiQ(iu) = iQ(maxL(1)-1,iu)      ! = Q(t)^(-1) at the optimal LS point
                if (optV(iu) < neginfty+tol) then
                    write(*,*) 'no possible choice of employment is affordable for single', g, a, s
                    stop
                end if
            end do
            
            ! integrate preference shocks
            EV(ie)  = sum(optV)*wg_uS
            EiQ(ie) = sum(optiQ)*wg_uS

        end do


        ! Integrate productivity conditional on productivity at a-1
        do ie_1=1,ngpe
            mtxI(a)%S(g,ieta,ih,ik_1,ie_1) = dot_product( TMe(g,s,a,ie_1,:), EV )
            iEiQ(a)%S(g,ieta,ih,ik_1,ie_1) = 1.0_wp / dot_product( TMe(g,s,a,ie_1,:), EiQ )
        end do

    end subroutine slnS_a



    ! 
    ! Singles: solution at age maxa
    !
    subroutine slnS_maxa(g)
        implicit none
        integer, intent(in) :: g
        integer  :: a, ieta, ith, s, ik_1, ih, L, ie, ie_1, iu, maxL(1)
        real(wp) :: w, resK0, k_1, alpha, Q(0:1,ngpu), optV(ngpu), optiQ(ngpu), EV(ngpe), EiQ(ngpe)

                        
        ! Initialise storage matrices 
        forall (a=mina:maxa)
            iEiQ(a)%S(g,:,:,:,:) = 0.0_wp
            mtxI(a)%S(g,:,:,:,:) = 0.0_wp
        end forall


        ! Calculate continuation value at maxa conditional on savings at the end of period k(maxa)=0
        a = maxa
        do ieta = 1,neta
        do ith  = 1,ntheta
        do s    = 1,ns
        do ik_1 = 1,ngpk

            ih  = getH(ith,s)
            k_1 = gKS(ik_1,s,ith,a-1,g)
            
            do ie = 1,ngpe

                ! calculate earnings rate
                w = earnings(g, s, lnyS(g,ith,s), a, ge(g,s,ie))

                ! calculate value of each possible LS point
                do L=0,1
                    resK0 = resourcesS (w, L, k_1, 0.0_wp)
                    do iu = 1,ngpu
                        alpha   = real(L,wp)*utS(g,ieta,s,a,gu(g,s,iu))
                        Q(L,iu) = (resK0 + alpha) * i2pQ
                    end do
                end do

                ! now select optimal LS and integrate taste shocks
                do iu = 1,ngpu
                    maxL = maxloc(Q(:,iu))
                    optV(iu)  = 2.0_wp * log(Q(maxL(1)-1,iu)) + lnpQ
                    optiQ(iu) = 1.0_wp / Q(maxL(1)-1,iu)
                end do
                EV(ie)  = sum(optV)*wg_uS
                EiQ(ie) = sum(optiQ)*wg_uS
                
            end do
            
            ! integrate productivity conditional on productivity at a-1
            do ie_1=1,ngpe
                mtxI(a)%S(g,ieta,ih,ik_1,ie_1) = dot_product( TMe(g,s,a,ie_1,:), EV )
                iEiQ(a)%S(g,ieta,ih,ik_1,ie_1) = 1.0_wp / dot_product( TMe(g,s,a,ie_1,:), EiQ )
            end do

        end do
        end do
        end do
        end do

    end subroutine slnS_maxa




    ! ----------------------------------------------------------------------------------------------------------------
    ! Working functions for couples and singles
    !
    
    !
    ! solves EE and computes optimal value and marginal utility conditional on labour supply
    !
    subroutine solveEE(gk, alpha, Kmin, Kmax, resK0, iEiQ1, I1, k, V, iQ)
        implicit none 
        real(wp), intent(in)  :: gk(ngpk), alpha, Kmin, Kmax, resK0, iEiQ1(ngpk), I1(ngpk)
        real(wp), intent(out) :: k, V, iQ
        real(wp)              :: lk, uk, I1lk, I1uk, eulerArr(ngpk)
        integer               :: ik, iuk, ilk


        ! look for solution in interval [Kmin,Kmax]
        if (Kmin+tol<Kmax) then

            ! initialise Euler expression to zero
            eulerArr = 0.0_wp
        
            ! calculate value of euler expression at affordable savings levels K(a): eulerArr
            !     this function is decreasing with K(a) and optimal K(a) is its zero
            do ik = 1,ngpk
            if (ik<ngpk .and. gk(ik)<Kmax) then

                eulerArr(ik) = (resK0-gk(ik)+alpha)*i2pQ - iBR*iEiQ1(ik)
                if (eulerArr(ik)<=0.0_wp) then
                    iuk = ik;           uk = gK(iuk);  I1uk = I1(iuk)
                    ilk = max(ik-1,1);  lk = gK(ilk);  I1lk = I1(ilk)
                    exit
                end if

            else
            
                ilk = max(ik-1,1);  lk = gK(ilk);  I1lk = I1(ilk)
                iuk = ik;           uk = Kmax;     I1uk = LinInterp1d (Kmax, gK(ilk), gK(iuk), I1(ilk), I1(iuk))
                eulerArr(ik) = (resK0-Kmax+alpha)*i2pQ - iBR*LinInterp1D(Kmax, gK(ilk), gK(iuk), iEiQ1(ilk), iEiQ1(iuk))
                exit
            
            end if
            end do

            ! solve euler equation to get optimal savings conditional on LS
            !     solution is in interval [gk(lbk),gk(ubk)]
            if (ilk==iuk) then                                                              ! same lower and upper bound: at mink
                k = lk
            else
                if (eulerARR(ilk)*eulerARR(iuk)>0.0_wp .and. eulerArr(iuk)>0.0_wp) then     ! corner solution: would like to save more but need to consume cmin
                    k = uk
                else if (eulerARR(ilk)*eulerARR(iuk) <= 0.0_wp) then                        ! interior solution
                    k = LinInterp1D(0.0_wp, eulerARR(ilk), eulerARR(iuk), lk, uk)
                    if (k<Kmin) k = Kmin
                    if (k>Kmax) k = Kmax
                else
                    write(*,*) 'error: EE trying to find solution where it does not exist'
                    write(*,*) ilk, iuk, eulerARR(ilk), eulerARR(iuk)
                    stop
                end if
            end if

            ! calculate value and marginal utility at optimal savings        
            call getVdU(resK0, k, lk, uk, I1lk, I1uk, alpha, V, iQ)


        ! if only affordable savings is K=Kmin
        else if (abs(Kmax-Kmin)<tol) then

            k = gk(1)
            call getVdU(resK0, k, k, k, I1(1), I1(1), alpha, V, iQ)


        ! if Kmin>Kmax: impossible level of labour supply given earnings and assets
        else
        
            V  = neginfty
            iQ = 0.0_wp

        end if
            
    end subroutine solveEE


    !
    ! calculates optimal value and marginal utility conditional on labour supply
    !
    subroutine getVdU(resK0, k, lk, uk, I1lk, I1uk, alpha, V, iQ)
        implicit none
        real(wp), intent(in)  :: resK0, k, lk, uk, I1lk, I1uk, alpha
        real(wp), intent(out) :: V, iQ
        real(wp)              :: Q
                        
        Q  = (resK0 - k + alpha) * i2pQ
        V  = 2.0_wp*log(Q) + lnpQ + beta*LinInterp1D(k, lk, uk, I1lk, I1uk)
        iQ = 1.0_wp / Q
        
    end subroutine getVdU

end module solutionWL


