! --------------------------------------------------------------------------------------------------------------------
! This file is part of the package EMar, developed to produce the results in the paper "The marriage market, labor 
!    supply and education choice," by P.A. Chiapppori, M. Costa-Dias and C. Meghir. 
! Copyright (c) 2017 Monica Costa Dias
!
! Main routine: calls other routines to solve and simulate life-cycle problem in a equilibrum model of
!    the marriage market, estimate the model parameters and market prices, and run policy experiments.
! --------------------------------------------------------------------------------------------------------------------


program main

    use paths
    use precision,      only : wp
    use parameters,     only : paramS, paramWL, nParamMax
    use momData,        only : momWL_t
    use IOroutines,     only : paramLoadStarting, loadMatchByS, loadDataMomWL
    use simulationWL,   only : SolveAndSimWL
    use momentsWL,      only : SimMomentsWL
    use estimationWL,   only : estimateWL, calcVarianceWL
    use solutionSpH,    only : SolveAndSimSpH
    use estimationSpH,  only : estimateSpH, bootstrapSpH
    use experiments,    only : EducSubsidy, sensitivityrisk
    use omp_lib 
    use statamod

    implicit none
    
    integer                  :: nproc, proceed
    real(wp)                 :: Qform, V(nParamMax,nParamMax)
    type(momWL_t)            :: mom
    character(*), parameter  :: paramFileWL = paramPath//'initParamWL.txt'    ! parameters in post-marriage problem
    character(*), parameter  :: paramFileS  = paramPath//'initParamS.txt'     ! parameters in education and marriage problem



    ! Set nr of parallel processors
    nproc = 6
    call omp_set_num_threads(nproc) 
    !$omp parallel 
    if (omp_get_thread_num() == 0) write(*,'(a,i1)') 'Number of threads: ', omp_get_num_threads() 
    !$omp end parallel 

    ! Read initial information (parameters and data moments) 
    call paramLoadStarting(paramWL, paramFileWL)     ! Load parameters in post-marriage problem
    call paramLoadStarting(paramS,  paramFileS)      ! Load parameters in education & sorting problem
    call loadMatchByS                                ! Load matching patterns by education (data)


    !-----------------------------------------------------------------------------------------------------------------
    ! Solve post-marriage life-cycle problem (working life for couples and singles)
    write(*, '(/, "Press 1 to solve and simulate the WL problem, or press 0 otherwise  ")', ADVANCE='NO')
    read(*,*) proceed
    if (proceed==1) then
        call SolveAndSimWL(1)                                                   ! Solve LC problem and simulate LC profiles
        call SimMomentsWL(.true., momPath//'SimMomWL.txt', Qform, mom)          ! simulate moments
        write( *, '(/,"    Sum of Squared Differences in Moments: ", F15.6)') Qform
        write( *, '(  "    Solution and simulation of WL problem completed" )' )
        write(*,  '(/,"---------------------------------------------------------------------------",/)') 
    end if


    !-----------------------------------------------------------------------------------------------------------------
    ! Estimate parameters in post-marriage problem
    write(*, '(/,"Press 1 to estimate parameters in WL problem, or press 0 otherwise  ")', ADVANCE='NO')
    read(*,*) proceed
    if (proceed==1) then
        call estimateWL
        write( *, '( "    Estimation of parameters in WL problem completed" )' )
        write(*,  '(/,"---------------------------------------------------------------------------",/)') 
    end if
    

    !-----------------------------------------------------------------------------------------------------------------
    ! Variance of estimated parameters in 3rd stage problem
    write(*,*)
    write(*, '("Press 1 to estimate the variance of the parameters in the WL problem, or press 0 otherwise  ")', ADVANCE='NO')
    read(*,*) proceed
    if (proceed==1) then
        call calcVarianceWL(paramWL%ntoestimate, V(1:paramWL%ntoestimate,1:paramWL%ntoestimate))
        write( *, '( "    Estimation of the covariance matrix of the estimates of WL parameters completed" )' )
        write(*,  '(/,"---------------------------------------------------------------------------",/)') 
    end if
    

    !-----------------------------------------------------------------------------------------------------------------
    ! Solve and simulate education and sorting problem
    write(*,*)
    write(*, '("Press 1 to solve and simulate the SPH problem, or press 0 otherwise  ")', ADVANCE='NO')
    read(*,*) proceed
    if (proceed==1) then
        call SolveAndSimSpH(.true., resultspath//'resultsSpH.txt', .true., mompath//'SimMomSpH.txt', 1, Qform)
        write( *, '(/,"    Sum of Squared Differences in Moments: ", F15.6)') Qform
        write( *, '(  "    Solution and simulation of the SpH problem completed" )' )
        write(*,  '(/,"---------------------------------------------------------------------------",/)') 
    end if                                     


    !-----------------------------------------------------------------------------------------------------------------
    ! Estimate parameters in education and marital sorting
    write(*,*)
    write(*, '("Press 1 to estimate the parameters in the SpH problem, or press 0 otherwise  ")', ADVANCE='NO')
    read(*,*) proceed
    if (proceed==1) then 
        call estimateSpH (1, 1, Qform)
        write( *, '( "    Estimation of the parameters in the SpH problem completed" )' )
        write(*,  '(/,"---------------------------------------------------------------------------",/)') 
    end if
        


    !-----------------------------------------------------------------------------------------------------------------
    ! Calculate SE of estimated parameters in education and marital sorting
    write(*,*)
    write(*, '("Press 1 to calculate the SE of estimated parameters in the SpH problem, or press 0 otherwise  ")', ADVANCE='NO')
    read(*,*) proceed
    if (proceed==1) then 
        call bootstrapSpH
        write( *, '( "    Estimation of the SE of the parameters in the SpH problem completed" )' )
        write(*,  '(/,"---------------------------------------------------------------------------",/)') 
    end if
    
    
    !-----------------------------------------------------------------------------------------------------------------
    ! Simulate effect of changes in cost of education
    write(*,*)
    write(*, '("Press 1 to simulate the impact of subsidising University education, or press 0 otherwise  ")', ADVANCE='NO')
    read(*,*) proceed
    if (proceed==1) then
        call EducSubsidy
        write( *, '( "    Impact of education subsidy simulated" )' )
        write(*,  '(/,"---------------------------------------------------------------------------",/)') 
    end if


    !-----------------------------------------------------------------------------------------------------------------
    ! Simulate effects of changing risk
    write(*,*)
    write(*, '("Press 1 to simulate sensitivity to risk, or press 0 otherwise  ")', ADVANCE='NO')
    read(*,*) proceed
    write(*,*)
    if (proceed==1) then 
        call sensitivityrisk
        write( *, '( "    Simulations of reactions to risk completed" )' )
        write(*,  '(/,"---------------------------------------------------------------------------",/)') 
    end if

end program main

    