** -----------------------------------------------------------------------------
** This file draws randomly from the joint distribution of WL parameters and 
** stores the draws in "moments\rd-paramWL.dta". These draws are an input to 
** estimate variance of the parameters in the SpH problem.
** 
** Output file needs to be saved in STATA 11 to be readable by STATAMOD.
** -----------------------------------------------------------------------------



** Random draws from joint distribution of WL parameters
** -----------------------------------------------------------------------------

drop _all
clear matrix
cd "C:\EMar"
set matsize 11000


** read estimated parameters
infix 2 first coef 16-27 using parameters\estimatesWL.txt
mkmat coef, matrix (b)


** read covariance matrix
use "parameters\CovWL.dta", clear
mkmat V1-V125, matrix(V)


** truncation points
drop _all
global N=125
set obs $N
gen n=_n

gen truncU=0
replace truncU=1 if inrange(n,21,26) | inrange(n,87,125)                        // AR coeffs and density of UH
gen truncL=0
replace truncL=1 if inrange(n,3,26) | inrange(n,39,44) | inrange(n,81,125)      // AR coeffs, distr of UH, variances
global ub = 0.9999999
global lb = 0.0000001


** variances of non-truncated distribution
forvalues i=1/$N {
if (truncL[`i']==1) {
	
	local oldv = 0
	local oldb = 0
	local v = V[`i',`i']
	local b = b[`i',1]

	while ( abs((`v'-`oldv')/`v')>0.01) {
		
		local oldv = `v'
			
		local l    = ($lb-`b')/sqrt(`v')
		local phil = normalden(`l')
		local Phil = normal(`l')

		if (truncU[`i']==1) {
			local u    = ($ub-`b')/sqrt(`v')
			local phiu = normalden(`u')
			local Phiu = normal(`u')
		}
		else {
			local u    = 0
			local phiu = 0
			local Phiu = 1
		}

		local f1 = (`l'*`phil'-`u'*`phiu') / (`Phiu'-`Phil')
		local f2 = (`phil'-`phiu') / (`Phiu'-`Phil')
		local v  = V[`i',`i'] / ( 1 + `f1' - `f2'*`f2' )
		
	}

	mat V[`i',`i'] = `v'

}
}


** cholesky decomposition of V
mat L = cholesky(V)


** Draw from truncated joint normal 

set seed 372807
global N=500
local n=0
local i=0
while (`n'<$N) {

	local i=`i'+1
	
	gen sn = rnormal()
	mkmat sn, matrix(sn)
	mat rN = b + L*sn
	
	local keep = 1
	forvalues j=1/125 {
	if ((truncU[`j']==1 & rN[`j',1]>$ub) | (truncL[`j']==1 & rN[`j',1]<$lb)) {
		local keep = 0 
	}
	}
	
	if (`keep'==1) {
		local n=`n'+1
		svmat rN
		rename rN1 r`n'
		di `i', `n'
	}
	
	drop sn

}

saveold "moments\rd-paramWL.dta", replace version(11)






