** -----------------------------------------------------------------------------
** This file estimates the set of moments used in estimating the parameters 
** in the SpH problem and their standard errors.
** 
** It then stores the first 100 draws from the distribution of these moments to 
** feed the bootstrapping routine that estimates the standard errors of the 
** estimated parameters.
**
** It also stores random draws from the sorting matrix by education in 
** "moments\rd-Psmsf.dta", needed to estimate variance of the parameters 
** sorting patterns by HC.
**
** Both .dta output files need to be saved in STATA 11 to be readable by 
** STATAMOD.
** -----------------------------------------------------------------------------


drop _all
clear matrix
cd "C:\EMar"



** Routines
** -----------------------------------------------------------------------------

cap program drop calcmoments
program define calcmoments

	local i = $iter

	* proportion marrying
	foreach g in m w {
		sum couple if `g'==1 & `g'_insampleWL==1
		local `g'mean = r(mean)
	}
	global pmarry = 0.5*(`mmean'+`wmean')
	di $pmarry
	global pC = $pmarry/(2-$pmarry)
	global pS = (1-$pmarry)/(2-$pmarry)

	
	** joint distribution of education, gender and marital status
	global N = 0
	forvalues ws=1/3 {
	forvalues ms=1/3 {
		count if wS==`ws' & mS==`ms' & m_insampleWL==1 & w_insampleWL==1
		global Nw`ws'm`ms' = r(N)
		global N = $N+${Nw`ws'm`ms'}
	}
	}

	forvalues ws=1/3 {
	forvalues ms=1/3 {
		global pw`ws'm`ms'_`i' = ${Nw`ws'm`ms'}*${pC}/$N
	}
	}
		
	global Nw=0
	global Nm=0
	forvalues s=1/3 {
		count if ws==`s' & couple==0
		global Nw`s'=r(N)
		global Nw = $Nw+${Nw`s'}
		
		count if ms==`s' & couple==0
		global Nm`s'=r(N)
		global Nm = $Nm+${Nm`s'}
	}
	forvalues s=1/3 {
		global pw`s'm0_`i'=${Nw`s'}*${pS}/${Nw}
		global pw0m`s'_`i'=${Nm`s'}*${pS}/${Nm}
	}

	global pw0m0_`i' = 0
	
	
	* distribution of education, by gender
	forvalues s=1/3 {
		global pm`s'_`i' = (${Nm`s'}+${Nw1m`s'}+${Nw2m`s'}+${Nw3m`s'})/($N+$Nm)
		global pw`s'_`i' = (${Nw`s'}+${Nw`s'm1}+${Nw`s'm2}+${Nw`s'm3})/($N+$Nw)
	}


	** regression of education on (pYres,ct), by gender
	foreach g in m w {
		cap drop s1 s2 s3
		ta `g's23 if `g'==1 & `g'_insampleEduc==1, g(s)
		gen `g'pYres2 = `g'pYres^2
		forvalues s=1/3 {
			if (`i'==0) eststo `g'`s': regress s`s' `g'pYres `g'pYres2          ///
			                               if `g'==1 & `g'_insampleEduc==1
			if (`i'~=0)                regress s`s' `g'pYres `g'pYres2          ///
			                               if `g'==1 & `g'_insampleEduc==1
			global r`g'`s'_ct_`i'  = _b[_cons]
			global r`g'`s'_pY_`i'  = _b[`g'pYres]
			global r`g'`s'_pY2_`i' = _b[`g'pYres2]
		}
	}

end



** Estimate moments and their standard errors
** -----------------------------------------------------------------------------

** estimate moments
global iter = 0
eststo clear
use "data\BHPSsample.dta", clear
qui calcmoments

** bootstrap st errors
global Nbs = 5000
set seed 8705871
forvalues j=1/$Nbs {
	global iter =`j'
	use "data\BHPSsample.dta", clear
	qui bsample
	qui calcmoments
	di "BS iteration ", `j'
}

** save estimates and se
qui drop _all
qui set obs $Nbs
gen n=_n

forvalues ms=0/3 {
forvalues ws=0/3 {
	qui gen pw`ws'm`ms'=.
	forvalues i=1/$Nbs {
		qui replace pw`ws'm`ms' = ${pw`ws'm`ms'_`i'} if n==`i'
	}
	qui sum pw`ws'm`ms'
	global SEpw`ws'm`ms' = r(sd)
}
}

foreach g in m w {
forvalues s=1/3 {

	qui gen p`g'`s'=.
	forvalues i=1/$Nbs {
		qui replace p`g'`s' = ${p`g'`s'_`i'} if n==`i'
	}
	qui sum p`g'`s'
	global SEp`g'`s' = r(sd)
}
}

global vlist "ct pY"
foreach g in m w {
forvalues s=1/3 {
foreach v of global vlist {
	qui gen r`g'`s'_`v'=.
	forvalues i=1/$Nbs {
		qui replace r`g'`s'_`v' = ${r`g'`s'_`v'_`i'} if n==`i'
	}
	qui sum r`g'`s'_`v'
	global SEr`g'`s'_`v' = r(sd)
}
}
}

save "data\b_momentsSpH.dta", replace



** export data moments
** -----------------------------------------------------------------------------

forvalues ms=0/3 {
	matrix input Pwsm`ms' = (${pw1m`ms'_0}, ${SEpw1m`ms'},                      ///
	                         ${pw2m`ms'_0}, ${SEpw2m`ms'},                      ///
							 ${pw3m`ms'_0}, ${SEpw3m`ms'})
	matrix rownames Pwsm`ms' = "ms=`ms', ws=1-3"
}
matrix input Pw0ms = (${pw0m1_0}, ${SEpw0m1}, ${pw0m2_0}, ${SEpw0m2},           ///
                      ${pw0m3_0}, ${SEpw0m3})
matrix rownames Pw0ms = "ms=1-3, ws=0"

foreach g in m w {
	matrix input P`g's = (${p`g'1_0}, ${SEp`g'1}, ${p`g'2_0}, ${SEp`g'2},       ///
	                      ${p`g'3_0}, ${SEp`g'3})
	matrix rownames P`g's = "g=`g', s=1-3"
}

estout matrix(Pwsm0, fmt(5))using "moments\probSmSw.txt", replace style(fixed)  ///
	mlabels(none) collabels(none) title("! joint distribution (Sm,Sw)")
forvalues sm=1/3 {
	estout matrix(Pwsm`sm', fmt(5))using "moments\probSmSw.txt", append         ///
		style(fixed) mlabels(none) collabels(none)
}
estout matrix(Pw0ms, fmt(5))using "moments\probSmSw.txt", append style(fixed)   ///
	mlabels(none) collabels(none)

foreach g in m w {
	estout matrix(P`g's, fmt(5)) using "moments\probS`g'.txt", replace          ///
	style(fixed) mlabels(none) collabels(none)                                  ///
	title("! distribution S by factors: `g'")
}

estout m1 m2 m3 using "moments\regrSm.txt", replace style(fixed) label          ///
	cells("b(fmt(3)) se(fmt(3))") mlabels(secondary further higher)             ///
	varlabels(_cons "intercept" mpYres "residual pY" ) order(_cons mpYres)
	
estout w1 w2 w3 using "moments\regrSw.txt", replace style(fixed) label          ///
	cells("b(fmt(3)) se(fmt(3))") mlabels(secondary further higher)             ///
	varlabels(_cons "intercept" wpYres "residual pY" ) order(_cons wpYres)



** Subsample of moments to bootstrap SE of parmeters in SpH problem by
** -----------------------------------------------------------------------------
use "moments\b_momentsSpH.dta", replace
keep in 1/100
saveold "moments\b100_momentsSpH.dta", replace version(11)



** Random draws from distribution of matching patterns by education
** To estimate covariance matrix of matching patterns by human capital
** -----------------------------------------------------------------------------
use "moments\b_momentsSpH.dta", replace
keep pw0m0-pw3m3
keep in 1/500
mkmat pw0m0-pw3m3, mat(A)
mat P = A'
drop _all
svmat P
saveold "moments\rd-Psmsf.dta", replace version(11)



** Delete file of bootstrap replications
** -----------------------------------------------------------------------------
erase "moments\b_momentsSpH.dta"


